import logging

import numpy as np


from .typeutils import (
    get_a_buffer,
    Info,
    is_integer,
    _to_slot_size,
    default_conf,
)
from .scalar import Int64, is_scalar

log = logging.getLogger(__name__)


"""
array itemtype d1 d2 ...
array itemtype d1 : ...
array itemtype (d1,1) (d1,0) ...  F contiguos

There 6 kind of arrays from the combination of
    shape: static, dynamic
    item: static, dynamic


Data layout:
    - [size]: if not (static,static)
    - [d0 d1 ...]: dynamic dimensions (dynamic,*)
    - [stride1 stride2 ...] if nd>1
    - [offsets]: if itemtype is not static (*|dynamic)
    - data: array data

Array class:
    - _size
    - _shape: the shape in the index space
    - _dshape_idx: index of dynamic shapes
    - _order: the hierarchical order of the indexes
    - _itemtype
    - _is_static_shape
    - _strides
    - _is_static_type

Array instance:
    - _dshape: value of dynamic dimensions
    - _shape: present if dynamic
    - _strides: shape if dynamic

"""


def get_suffix(shape):
    # write suffix  6x6x6 or Nx6xM
    sshape = []
    ilst = 0
    lst = "NMOPQRSTUVWXYZABCDEFGHIJKLM"
    # lst = "nmopqrstuvwxyzabcdefghijklm"
    for dd in shape:
        if dd is None:
            sshape.append(lst[ilst])
            ilst = (ilst + 1) % len(lst)
        else:
            sshape.append(str(dd))
    return "x".join(sshape)


def get_shape_from_array(value):
    if hasattr(value, "shape"):
        return value.shape
    elif hasattr(value, "_shape"):
        return value._shape
    if hasattr(value, "__len__"):
        shape = (len(value),)
        if len(value) > 0:
            shape0 = get_shape_from_array(value[0])
            if shape0 == ():
                return shape
            for i in value[1:]:
                shapei = get_shape_from_array(i)
                if shapei != shape0:
                    raise ValueError(f"{value} not an array")
            return shape + shape0
        else:
            return shape
    else:
        return ()


def get_strides(shape, order, itemsize):
    """
    shape dimension for each index
    order of the dimensions in the memory layout
    - 0 is slowest variation
    return strides for each index

    off=strides[0]*idx[0]+strides[1]*idx[1]+...+strides[n]*idx[n]

    """
    cshape = [shape[io] for io in order]
    cstrides = get_c_strides(cshape, itemsize)
    return tuple(cstrides[order.index(ii)] for ii in range(len(order)))


def get_f_strides(shape, itemsize):
    """
    calculate strides assuming F ordering
    """
    ss = itemsize
    strides = []
    for sh in shape:
        strides.append(ss)
        ss *= sh
    return tuple(strides)


def get_c_strides(shape, itemsize):
    """
    calculate strides assuming C ordering
    """
    ss = itemsize
    strides = []
    for sh in reversed(shape):
        strides.append(ss)
        ss *= sh
    return tuple(reversed(strides))


def iter_index(shape, order):
    """return index in order of data layout"""
    if len(shape) == 1:
        for ii in range(shape[0]):
            yield ii
    else:
        aorder = [order.index(ii) for ii in range(len(order))]
        for ii in np.ndindex(*[shape[io] for io in order]):
            yield tuple(ii[io] for io in aorder)


def mk_order(order, shape):
    if order == "C":
        return list(range(len(shape)))
    elif order == "F":
        return list(range(len(shape) - 1, -1, -1))
    else:
        return order


def get_offset(idx, strides):
    return sum(ii * ss for ii, ss in zip(idx, strides))


def bound_check(index, shape):
    for ii, ss in zip(index, shape):
        if ii < 0 or ii >= ss:
            raise IndexError(f"index {index} outside shape {shape}")


class Index:
    def __init__(self, cls):
        self.cls = cls


class MetaArray(type):
    def __new__(cls, name, bases, data):
        if "_itemtype" in data:  # specialized class
            _data_offset = 0
            _itemtype = data["_itemtype"]
            if _itemtype._size is None:
                static_size = 8
                data["_is_static_type"] = False
            else:
                data["_is_static_type"] = True
                static_size = _itemtype._size
            if "_shape" not in data:
                raise ValueError(f"No shape defined for {cls}")
            if "_order" not in data:
                data["_order"] = "C"
            _shape = data["_shape"]
            dshape = []  # find dynamic shapes
            for ii, d in enumerate(_shape):
                if d is None:
                    data["_is_static_shape"] = False
                    dshape.append(ii)
            if len(dshape) > 0:
                data["_is_static_shape"] = False
                data["_dshape_idx"] = dshape
                _data_offset += len(dshape) * 8  # space for dynamic shapes
                if len(_shape) > 1:
                    _data_offset += len(_shape) * 8  # space for strides
                else:
                    data["_strides"] = (static_size,)
            else:
                data["_is_static_shape"] = True
                data["_order"] = mk_order(data["_order"], _shape)
                data["_strides"] = get_strides(
                    _shape, data["_order"], static_size
                )

            if data["_is_static_shape"] and data["_is_static_type"]:
                _size = _itemtype._size
                for d in _shape:
                    _size *= d
                _size = _to_slot_size(_size)
            else:
                _size = None
                _data_offset += 8  # space for dynamic size

            data["_size"] = _size
            data["_data_offset"] = _data_offset
        # need to applied to derived classes as well
        if "_c_type" not in data:
            data["_c_type"] = name

        return type.__new__(cls, name, bases, data)

    def __getitem__(cls, shape):
        return Array.mk_arrayclass(cls, shape)

    def _get_offset(cls, index):
        return get_offset(index, cls._strides)

    def _get_n_items(cls):
        if cls._is_static_shape:
            return np.prod(cls._shape)
        else:
            raise ValueError("Cannot get n items from dynamic shapes")

    def __repr__(cls):
        return f"<array {cls.__name__}>"


class Array(metaclass=MetaArray):
    _shape: tuple
    _order: tuple
    _strides: tuple
    _itemtype: type
    _size: int
    _dshape: tuple
    _is_static_shape: bool
    _is_static_type: bool
    _data_offset: int

    @classmethod
    def mk_arrayclass(cls, itemtype, shape):
        if type(shape) in (int, slice):
            shape = (shape,)
        order = list(range(len(shape)))
        nshape = []
        for ii, dd in enumerate(shape):
            if type(dd) is slice:
                nshape.append(dd.start)
                if dd.stop is not None:
                    order[ii] = dd.stop
            else:
                nshape.append(dd)

        suffix = get_suffix(nshape)

        name = f"Arr{suffix}{itemtype.__name__}"

        data = {
            "_itemtype": itemtype,
            "_shape": tuple(nshape),
            "_order": tuple(order),
        }
        return MetaArray(name, (cls,), data)

    @classmethod
    def _inspect_args(cls, *args):
        """
        Determine:
        - size:
        - shape, order, strides
        - offsets
        - value: None if args contains dimensions else args[0]
        """
        log.debug(f"get size for {cls} from {args}")
        info = Info()
        extra = {}
        if cls._size is not None:
            # static,static array
            if len(args) == 0:
                value = None
            elif len(args) == 1:
                shape = get_shape_from_array(args[0])
                if shape != cls._shape:
                    raise ValueError(f"shape not valid for {args[0]} ")
                value = args[0]
            elif len(args) > 1:
                raise ValueError("too many arguments")
            size = cls._size
            shape = cls._shape
            order = cls._order
            strides = cls._strides
            offsets = np.empty(shape, dtype="int64")
            offset = 0
            for idx in iter_index(shape, order):
                offsets[idx] = offset
                offset += cls._itemtype._size
            assert size == _to_slot_size(offset)
        else:  # handling other cases
            offset = 8  # space for size data
            # determine shape and order
            if cls._is_static_shape:
                shape = cls._shape
                order = cls._order
                strides = cls._strides
                items = np.prod(shape)
                value = args[0]
            else:  # complete dimensions
                if not is_integer(args[0]):  # init with array
                    value = args[0]
                    shape = get_shape_from_array(value)
                    dshape = []
                    for idim, ndim in enumerate(cls._shape):
                        if ndim is None:
                            dshape.append(idim)
                        else:
                            if shape[idim] != ndim:
                                raise ValueError(
                                    "Array: incompatible dimensions"
                                )
                    value = value
                else:  # init with shapes
                    value = None
                    shape = []
                    dshape = []  # index of dynamic shapes
                    for ndim in cls._shape:
                        if ndim is None:
                            shape.append(args[len(dshape)])
                            dshape.append(len(shape))
                        else:
                            shape.append(ndim)
                # now we have shape, dshape
                shape = shape
                info.dshape = dshape
                offset += len(dshape) * 8  # space for dynamic shapes
                if len(shape) > 1:
                    offset += len(shape) * 8  # space for strides
                order = mk_order(cls._order, shape)
                if cls._is_static_type:
                    strides = get_strides(shape, order, cls._itemtype._size)
                else:
                    strides = get_strides(shape, order, 8)
                items = np.prod(shape)

            # needs items, order, shape, value
            if cls._is_static_type:
                # offsets = np.empty(shape, dtype="int64")
                # for idx in iter_index(shape, order):
                #    offsets[idx] = offset
                #    offset += cls._itemtype._size
                offset += cls._itemtype._size * items
                size = _to_slot_size(offset)

            else:
                # args must be an array of correct dimensions
                offsets = np.empty(shape, dtype="int64")
                offset += items * 8
                for idx in iter_index(shape, order):
                    extra[idx] = cls._itemtype._inspect_args(value[idx])
                    offsets[idx] = offset
                    offset += extra[idx].size
                size = _to_slot_size(offset)
                info.offsets = offsets
                info.extra = extra

        info.shape = shape
        info.strides = strides
        info.size = size
        info.order = order
        info.value = value
        info.items = np.prod(shape)
        return info

    @classmethod
    def _from_buffer(cls, buffer, offset):
        self = object.__new__(cls)
        self._buffer = buffer
        self._offset = offset
        coffset = offset
        if cls._size is None:
            self._size = Int64._from_buffer(self._buffer, coffset)
            coffset += 8
        if not cls._is_static_shape:
            shape = []
            for dd in cls._shape:
                if dd is None:
                    shape.append(Int64._from_buffer(self._buffer, coffset))
                    coffset += 8
                else:
                    shape.append(dd)
            self._shape = shape
            if len(self._shape) > 1:  # getting strides
                # could be computed from shape and order but offset needs to taken
                strides = []
                for _ in range(shape):
                    strides.append(Int64._from_buffer(self._buffer, coffset))
                    coffset += 8
            else:
                if cls._is_static_type:
                    strides = (cls._itemtype._size,)
                else:
                    strides = (8,)
            self._strides = tuple(strides)
        else:
            shape = cls._shape
        if not cls._is_static_type:
            items = np.prod(shape)
            self._offsets = Int64._array_from_buffer(buffer, coffset, items)
        return self

    @classmethod
    def _to_buffer(cls, buffer, offset, value, info=None):
        if info is None:
            info = cls._inspect_args(value)
        value = info.value  # can be None if value contained shape info
        header = []
        coffset = offset
        if cls._size is None:
            header.append(info.size)
        if not cls._is_static_shape:
            for ii, nd in enumerate(cls._shape):
                if nd is None:
                    header.append(info.shape[ii])
            if len(cls._shape) > 1:
                header.extend(info.strides)
        if len(header) > 0:
            Int64._array_to_buffer(
                buffer, coffset, np.array(header, dtype="i8")
            )
            coffset += 8 * len(header)
        if not cls._is_static_type:
            Int64._array_to_buffer(buffer, coffset, info.offsets)
            coffset += 8 * len(info.offsets)
        if hasattr(cls._itemtype, "_dtype") and hasattr(
            value, "dtype"
        ):  # is a scalar type:
            if not isinstance(value, buffer.context.nplike_array_type):
                value = buffer.context.nparray_to_context_array(value)
            buffer.update_from_nplike(coffset, cls._itemtype._dtype, value)
        elif isinstance(value, cls):
            if value._size == info.size:
                buffer.update_from_xbuffer(
                    offset, value._buffer, value._offset, value._size
                )
            else:
                raise ValueError("Value {value} not compatible size")
        elif value is None:  # no value to initialize
            if is_scalar(cls._itemtype):
                pass  # leave uninitialized
            else:
                value = cls._itemtype()  # use default type
                if cls._is_static_type:
                    ioffset = offset + cls._data_offset
                    for idx in range(info.items):
                        cls._itemtype._to_buffer(buffer, ioffset, value, None)
                        ioffset += cls._itemtype._size
                else:
                    for idx in iter_index(info.shape, cls._order):
                        cls._itemtype._to_buffer(
                            buffer,
                            offset + info.offsets[idx],
                            value[idx],
                            info.extra.get(idx),
                        )
        else:  # there is a value for initialization
            if cls._is_static_type:
                ioffset = offset + cls._data_offset
                for idx in iter_index(info.shape, cls._order):
                    cls._itemtype._to_buffer(
                        buffer, ioffset, value[idx], info=None
                    )
                    ioffset += cls._itemtype._size
            else:
                for idx in iter_index(info.shape, cls._order):
                    cls._itemtype._to_buffer(
                        buffer,
                        offset + info.offsets[idx],
                        value[idx],
                        info.extra.get(idx),
                    )

    def __init__(self, *args, _context=None, _buffer=None, _offset=None):
        # determin resources
        cls = self.__class__
        info = cls._inspect_args(*args)

        self._buffer, self._offset = get_a_buffer(
            info.size, _context, _buffer, _offset
        )

        cls._to_buffer(self._buffer, self._offset, info.value, info)

        if cls._size is None:
            self._size = info.size
        if not cls._is_static_shape:
            self._shape = info.shape
            self._dshape = info.dshape
            self._strides = info.strides
        if not cls._is_static_type:
            self._offsets = info.offsets

    def _get_size(self):
        if self.__class__._size is None:
            return Int64._from_buffer(self._buffer, self._offset)
        else:
            return self.__class__._size

    @classmethod
    def _get_position(cls, index):
        offset = get_offset(index, cls._strides)
        if cls._is_static_type:
            return offset // 8
        else:
            return offset // cls._itemtype._size

    def __getitem__(self, index):
        if isinstance(index, (int, np.integer)):
            index = (index,)
        cls = self.__class__
        if hasattr(self, "_offsets"):
            offset = self._offset + self._offsets[index]
        else:
            bound_check(index, self._shape)
            offset = (
                self._offset
                + cls._data_offset
                + get_offset(index, self._strides)
            )
        return cls._itemtype._from_buffer(self._buffer, offset)

    def __setitem__(self, index, value):
        if isinstance(index, (int, np.integer)):
            index = (index,)
        cls = self.__class__
        if hasattr(cls._itemtype, "_update"):
            self[index]._update(value)
        else:
            if hasattr(self, "_offsets"):
                offset = self._offset + self._offsets[index]
            else:
                bound_check(index, self._shape)
                offset = (
                    self._offset
                    + cls._data_offset
                    + get_offset(index, self._strides)
                )
            cls._itemtype._to_buffer(self._buffer, offset, value)

    def _get_offset(self, index):
        if isinstance(index, (int, np.integer)):
            index = (index,)
        cls = self.__class__
        if hasattr(self, "_offsets"):
            offset = self._offset + self._offsets[index]
        else:
            bound_check(index, self._shape)
            offset = (
                self._offset
                + cls._data_offset
                + get_offset(index, self._strides)
            )
        return offset

    def _iter_index(self):
        return iter_index(self._shape, self._order)

    def __len__(self):
        return np.prod(self._shape)

    def to_nplike(self):
        shape = self._shape
        cshape = [shape[ii] for ii in self._order]
        if hasattr(self._itemtype, "_dtype"):
            arr = self._buffer.to_nplike(
                self._offset + self._data_offset, self._itemtype._dtype, cshape
            ).transpose(self._order)
            assert arr.strides == self._strides
            return arr
        else:
            raise NotImplementedError

    @classmethod
    def _gen_data_paths(cls, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [cls])
        path = base + [cls, Index(cls)]
        paths.append(path)
        if hasattr(cls._itemtype, "_gen_data_paths"):
            paths.extend(cls._itemtype._gen_data_paths(path))
        return paths

    @classmethod
    def _gen_c_api(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_code(cls, paths, conf)

    @classmethod
    def _gen_c_decl(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_cdefs(cls, paths, conf)

    @classmethod
    def _gen_kernels(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_kernels(cls, paths, conf)

    def __repr__(self):
        return f"<{self.__class__.__name__}{self._shape} at {self._offset}>"

    @classmethod
    def _get_inner_types(cls):
        return [cls._itemtype]


def is_index(atype):
    return isinstance(atype, Index)


def is_array(atype):
    return isinstance(atype, MetaArray)



"""
Scalars: Types olding numbers


NB: scalars cannot be classes as classes() needs to return a numpy scalar, however in union isintance could be used (try to subclass numpy scalar?)
"""

import numpy as np
import logging
from .typeutils import Info


log = logging.getLogger(__name__)


class NumpyScalar:
    def __init__(self, dtype, cname):
        self.__name__ = dtype.capitalize()
        self._dtype = np.dtype(dtype)
        self._size = self._dtype.itemsize
        self._c_type = cname

    def _from_buffer(self, buffer, offset):
        data = buffer.to_bytearray(offset, self._size)
        return np.frombuffer(data, dtype=self._dtype)[0]

    def _to_buffer(self, buffer, offset, value, info=None):
        data = self._dtype.type(value).tobytes()
        buffer.update_from_buffer(offset, data)

    def __call__(self, value=0):
        return self._dtype.type(value)

    def _inspect_args(self, arg):
        return Info(size=self._size)

    def __getitem__(self, shape):
        from .array import Array  # avoid circular import

        return Array.mk_arrayclass(self, shape)

    def __repr__(self):
        return self.__name__

    def _array_to_buffer(self, buffer, offset, value):
        return buffer.update_from_buffer(offset, value.tobytes())

    def _array_from_buffer(self, buffer, offset, count):
        return buffer.to_nplike(offset, self._dtype, (count,))

    def _gen_data_paths(self, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [self])
        return paths


Float64 = NumpyScalar("float64", "double")
Float32 = NumpyScalar("float32", "float")
Int64 = NumpyScalar("int64", "int64_t")
UInt64 = NumpyScalar("uint64", "uint64_t")
Int32 = NumpyScalar("int32", "int32_t")
UInt32 = NumpyScalar("uint32", "uint32_t")
Int16 = NumpyScalar("int16", "int16_t")
UInt16 = NumpyScalar("uint16", "uint16_t")
Int8 = NumpyScalar("int8", "int8_t")
UInt8 = NumpyScalar("uint8", "uint8_t")
Complex64 = NumpyScalar("complex64", "float[2]")
Complex128 = NumpyScalar("complex128", "double[2]")
#Complex256 = NumpyScalar("complex256", "double[4]")  # incompatible with M1 CPU
#Float128 = NumpyScalar("float128", "double[2]")      # incompatible with M1 CPU


def is_scalar(cls):
    return isinstance(cls, NumpyScalar)


class Void:
    _c_type = "void"
    _size = None




class BaseLinkedArray():

    container = None
    container_setitem_name=None
    mode = None

    @classmethod
    def from_array(cls, a, mode=None, container=None,
                   container_setitem_name=None):
        assert len(a.shape) == 1 # TODO: To be generalized
        assert mode in (None, 'readonly', 'setitem_from_container')
        self = cls._build_view(a)
        self.mode=mode
        self.container = container
        self.container_setitem_name = container_setitem_name
        return self

    def __setitem__(self, indx, val):
        if (self.mode is None or
            (hasattr(self.container, '_flag_bypass_linked')
                         and self.container._flag_bypass_linked)):
            super().__setitem__(indx, val)
        elif self.mode == 'setitem_from_container':
            getattr(self.container, self.container_setitem_name)(indx, val)
        elif self.mode == 'readonly':
            raise ValueError('This array is read only')

class BypassLinked:
    def __init__(self, container):
        self.container = container

    def __enter__(self):
        self.container._flag_bypass_linked = True

    def __exit__(self, *args, **kwargs):
        del(self.container._flag_bypass_linked)


import numpy as np

from .context_cpu import ContextCpu

context_default = ContextCpu()


def get_a_buffer(size, context=None, buffer=None, offset=None):
    if buffer is None:
        if offset is not None:
            raise ValueError("Cannot set `offset` without buffer")
        if context is None:
            context = context_default
        buffer = context.new_buffer(size)
    if offset is None:
        offset = buffer.allocate(size)
    elif offset == "aligned":
        offset = buffer.allocate(size, align=True)
    elif offset == "packed":
        offset = buffer.allocate(size, align=False)
    if isinstance(offset, str):
        raise ValueError(f"Invalid offset {offset}")
    return buffer, offset


def dispatch_arg(f, arg):
    if isinstance(arg, tuple):
        return f(*arg)
    elif isinstance(arg, dict):
        return f(**arg)
    else:
        return f(arg)


class Info:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        args = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        return f"Info({','.join(args)})"

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


def _to_slot_size(size):
    "round to nearest multiple of 8"
    return (size + 7) & (-8)


def _is_dynamic(cls):
    return cls._size is None


def is_integer(i):
    return isinstance(i, (int, np.integer))


float2c = {2: "half", 4: "float", 8: "double", 16: "double[2]"}


default_conf = {
    "gpumem": "/*gpuglmem*/",
    "cpurestrict": "/*restrict*/",
    "inttype": "int64_t",
    "chartype": "char",
    "gpufun": "/*gpufun*/",
}


def get_c_type(typ):
    if hasattr(typ, "dtype"):
        ss = typ.dtype.str
        tt = ss[1]
        nb = int(ss[2:])
        if tt == "f":
            return float2c[nb]
        elif tt == "i":
            return f"int{nb*8}_t"
        elif tt == "u":
            return f"int{nb*8}_t"
        elif tt == "c":
            return f"{float2c[nb//2]}[2]"
        elif tt == "S":
            return f"char[{nb}]"
    elif hasattr(typ, "_c_type"):
        return typ._c_type
    else:
        raise ValueError(f"Cannot find C type for type {typ}")


class Register:
    def __init__(self):
        self.classes = {}



from .scalar import (
    Float64,
    Float32,
    Int64,
    UInt64,
    Int32,
    UInt32,
    Int16,
    UInt16,
    Int8,
    UInt8,
)
from .array import Array
from .string import String
from .struct import Struct, Field
from .ref import Ref, UnionRef

from .context_cpu import ContextCpu
from .context_pyopencl import ContextPyopencl
from .context_cupy import ContextCupy

from .context import Arg, Kernel, Method, get_user_context

from .specialize_source import specialize_source

from .typeutils import context_default, get_a_buffer

from .dress import dress, JEncoder, DressedStruct

from .linkedarray import BypassLinked



import os
import logging

import numpy as np

from .context import (
    XBuffer,
    XContext,
    ModuleNotAvailable,
    available,
    classes_from_kernels,
    sort_classes,
    sources_from_classes,
    _concatenate_sources,
)

from .specialize_source import specialize_source
from .linkedarray import BaseLinkedArray

log = logging.getLogger(__name__)

try:
    import pyopencl as cl
    import pyopencl.array as cla

    _enabled = True
except ImportError:
    log.info(
        "pyopencl is not installed, ContextPyopencl will not be available"
    )
    cl = ModuleNotAvailable(
        message=(
            "pyopencl is not installed. " "ContextPyopencl is not available!"
        )
    )
    cl.Buffer = cl
    cla = cl
    _enabled = False

from ._patch_pyopencl_array import _patch_pyopencl_array


openclheader = [
    """\
#ifndef XOBJ_STDINT
typedef long int64_t;
typedef char int8_t;
typedef unsigned int uint32_t;
#endif"""
]

if _enabled:
    # order of base classes matters as it defines which __setitem__ is used
    class LinkedArrayPyopencl(BaseLinkedArray, cla.Array):

        @classmethod
        def _build_view(cls, a):
            assert len(a.shape) == 1
            return cls(cq=a.queue, shape=a.shape, dtype=a.dtype,
                    data=a.base_data, offset=a.offset, strides=a.strides, order='C',
                    _flags=a.flags)

class ContextPyopencl(XContext):
    @property
    def nplike_array_type(self):
        return cla.Array

    @property
    def linked_array_type(self):
        return LinkedArrayPyopencl

    @classmethod
    def get_devices(cls):
        out = []
        for ip, platform in enumerate(cl.get_platforms()):
            for id, device in enumerate(platform.get_devices()):
                out.append(f"{ip}.{id}")
        return out

    @classmethod
    def print_devices(cls):
        for ip, platform in enumerate(cl.get_platforms()):
            print(f"Context {ip}: {platform.name}")
            for id, device in enumerate(platform.get_devices()):
                print(f"Device {ip}.{id}: {device.name}")

    def __init__(
        self, device=None, patch_pyopencl_array=True, minimum_alignment=None
    ):

        """
        Creates a Pyopencl Context object, that allows performing the computations
        on GPUs and CPUs through PyOpenCL.

        Args:
            device (str or Device): The device (CPU or GPU) for the simulation.
            default_kernels (bool): If ``True``, the Xfields defult kernels are
                automatically imported.
            patch_pyopencl_array (bool): If ``True``, the PyOpecCL class is patched to
                allow some operations with non-contiguous arrays.
            specialize_code (bool): If True, the code is specialized using
                annotations in the source code. Default is ``True``

        Returns:
            ContextPyopencl: context object.

        """

        super().__init__()

        # TODO assume one device only
        if device is None:
            self.context = cl.create_some_context(interactive=False)
            self.device = self.context.devices[0]
            self.platform = self.device.platform
        else:
            if isinstance(device, str):
                platform, device = map(int, device.split("."))
            else:
                self.device = device
                self.platform = device.platform

            self.platform = cl.get_platforms()[platform]
            self.device = self.platform.get_devices()[device]
            self.context = cl.Context([self.device])

        self.queue = cl.CommandQueue(self.context)

        if patch_pyopencl_array:
            _patch_pyopencl_array(cl, cla, self.context)

        if minimum_alignment is None:
            minimum_alignment = self.find_minimum_alignment()
        self.minimum_alignment = minimum_alignment

    def _make_buffer(self, capacity):
        return BufferPyopencl(capacity=capacity, context=self)

    def find_minimum_alignment(self):
        buff = self.new_buffer()
        i = 1
        found = False
        while i < 2 ** 16:
            try:
                buff.buffer[i:]
                found = True
                break
            except cl._cl.RuntimeError:
                pass
            i += 1
        if not found:
            raise RuntimeError(
                "Impossible to find minimum alignment on Pyopencl context"
            )
        return i

    def add_kernels(
        self,
        sources=[],
        kernels=[],
        specialize=True,
        save_source_as=None,
        extra_cdef=None,
        extra_classes=[],
        extra_headers=[],
    ):

        """
        Adds user-defined kernels to to the context. The kernel source
        code is provided as a string and/or in source files and must contain
        the kernel names defined in the kernel descriptions.
        Args:
            sources (list): List of source codes that are concatenated before
                compilation. The list can contain strings (raw source code),
                File objects and Path objects.
            kernels (dict): Dictionary with the kernel descriptions
                in the form given by the following examples. The descriptions
                define the kernel names, the type and name of the arguments
                and identify one input argument that defines the number of
                threads to be launched (only on cuda/opencl).
            specialize_code (bool): If True, the code is specialized using
                annotations in the source code. Default is ``True``
            save_source_as (str): Filename for saving the specialized source
                code. Default is ```None```.
        Example:

        .. code-block:: python

            # A simple kernel
            src_code = '''
            /*gpukern*/
            void my_mul(const int n,
                /*gpuglmem*/ const double* x1,
                /*gpuglmem*/ const double* x2,
                /*gpuglmem*/       double* y) {
                int tid = 0 //vectorize_over tid
                y[tid] = x1[tid] * x2[tid];
                //end_vectorize
                }
            '''

            # Prepare description
            kernel_descriptions = {
                "my_mul": xo.Kernel(
                    args=[
                        xo.Arg(xo.Int32, name="n"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x1"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x2"),
                        xo.Arg(xo.Float64, pointer=True, const=False, name="y"),
                    ],
                    n_threads="n",
                    ),
            }

            # Import kernel in context
            ctx.add_kernels(
                sources=[src_code],
                kernels=kernel_descriptions,
                save_source_as=None,
            )

            # With a1, a2, b being arrays on the context, the kernel
            # can be called as follows:
            ctx.kernels.my_mul(n=len(a1), x1=a1, x2=a2, y=b)
        """

        classes = classes_from_kernels(kernels)
        classes.update(extra_classes)
        classes = sort_classes(classes)
        cls_sources = sources_from_classes(classes)

        headers = openclheader + extra_headers

        sources = headers + cls_sources + sources

        source, folders = _concatenate_sources(sources)

        if specialize:
            # included files are searched in the same folders od the src_filed
            specialized_source = specialize_source(
                source, specialize_for="opencl", search_in_folders=folders
            )
        else:
            specialized_source = source

        if save_source_as is not None:
            with open(save_source_as, "w") as fid:
                fid.write(specialized_source)

        prg = cl.Program(self.context, specialized_source).build()

        for pyname, kernel in kernels.items():
            if kernel.c_name is None:
                kernel.c_name = pyname

            self.kernels[pyname] = KernelPyopencl(
                function=getattr(prg, kernel.c_name),
                description=kernel,
                context=self,
            )

            self.kernels[pyname].source = source
            self.kernels[pyname].specialized_source = specialized_source

    def nparray_to_context_array(self, arr):

        """
        Copies a numpy array to the device memory.
        Args:
            arr (numpy.ndarray): Array to be transferred

        Returns:
            pyopencl.array.Array:The same array copied to the device.

        """
        dev_arr = cla.to_device(self.queue, arr)
        return dev_arr

    def nparray_from_context_array(self, dev_arr):

        """
        Copies an array to the device to a numpy array.

        Args:
            dev_arr (pyopencl.array.Array): Array to be transferred.
        Returns:
            numpy.ndarray: The same data copied to a numpy array.

        """
        return dev_arr.get()

    @property
    def nplike_lib(self):
        """
        Module containing all the numpy features supported by PyOpenCL (optionally
        with patches to operate with non-contiguous arrays).
        """
        return cla

    def synchronize(self):
        """
        Ensures that all computations submitted to the context are completed.
        No action is performed by this function in the Pyopencl context. The method
        is provided so that the Pyopencl context has an identical API to the Cupy one.
        """
        pass

    def zeros(self, *args, **kwargs):
        """
        Allocates an array of zeros on the device. The function has the same
        interface of numpy.zeros"""
        return self.nplike_lib.zeros(self.queue, *args, **kwargs)

    def plan_FFT(self, data, axes, wait_on_call=True):
        """
        Generates an FFT plan object to be executed on the context.

        Args:
            data (pyopencl.array.Array): Array having type and shape for which
                the FFT needs to be planned.
            axes (sequence of ints): Axes along which the FFT needs to be
                performed.
        Returns:
            FFTPyopencl: FFT plan for the required array shape, type and axes.

        Example:

        .. code-block:: python

            plan = context.plan_FFT(data, axes=(0,1))

            data2 = 2*data

            # Forward tranform (in place)
            plan.transform(data2)

            # Inverse tranform (in place)
            plan.itransform(data2)
        """
        return FFTPyopencl(self, data, axes, wait_on_call)

    @property
    def kernels(self):

        """
        Dictionary containing all the kernels that have been imported to the context.
        The syntax ``context.kernels.mykernel`` can also be used.
        """

        return self._kernels


class BufferPyopencl(XBuffer):
    def _make_context(self):
        return ContextPyopencl()

    def _new_buffer(self, capacity):
        return cl.Buffer(
            self.context.context, cl.mem_flags.READ_WRITE, capacity
        )

    def copy_from(self, source, src_offset, dest_offset, byte_count):
        # Does not pass through cpu if it can
        # source: python object that uses buffer protocol or opencl buffer
        cl.enqueue_copy(
            self.context.queue,
            self.buffer,
            source,
            src_offset,
            dest_offset,
            byte_count,
        )

    def write(self, offset, data):
        # From python object with buffer interface on cpu
        log.debug(f"write {self} {offset} {data}")
        cl.enqueue_copy(
            self.context.queue, self.buffer, data, device_offset=offset
        )

    def read(self, offset, size):
        # To bytearray on cpu
        data = bytearray(size)
        cl.enqueue_copy(
            self.context.queue, data, self.buffer, device_offset=offset
        )
        return data

    def update_from_native(
        self, offset: int, source: cl.Buffer, source_offset: int, nbytes: int
    ):
        """Copy data from native buffer into self.buffer starting from offset"""
        cl.enqueue_copy(
            self.context.queue,
            self.buffer,
            source,
            src_offset=source_offset,
            dest_offset=offset,
            byte_count=nbytes,
        )

    def copy_native(self, offset: int, nbytes: int):
        """return native data with content at from offset and nbytes"""
        buff = self._new_buffer(nbytes)
        cl.enqueue_copy(
            queue=self.context.queue,
            dest=buff,
            src=self.buffer,
            src_offset=offset,
            byte_count=nbytes,
        )
        return buff

    def copy_to_native(
        self, dest, dest_offset, source_offset: int, nbytes: int
    ):
        """return native data with content at from offset and nbytes"""
        cl.enqueue_copy(
            queue=self.context.queue,
            dest=dest,
            src=self.buffer,
            src_offset=source_offset,
            byte_count=nbytes,
        )

    def update_from_buffer(self, offset: int, source):
        """Copy data from python buffer such as bytearray, bytes, memoryview, numpy array.data"""
        cl.enqueue_copy(
            queue=self.context.queue,
            dest=self.buffer,
            src=source,  # nbytes taken from min(len(source),len(buffer))
            device_offset=offset,
        )

    def to_nplike(self, offset, dtype, shape):
        """view in nplike"""
        return cl.array.Array(
            self.context.queue,
            data=self.buffer,
            offset=offset,
            dtype=dtype,
            shape=tuple(shape),
        )

    def update_from_nplike(self, offset, dest_dtype, arr):
        if arr.dtype != dest_dtype:
            arr = arr.astype(dest_dtype)
        self.update_from_native(offset, arr.base_data, arr.offset, arr.nbytes)

    def to_bytearray(self, offset, nbytes):
        """copy in byte array: used in update_from_xbuffer"""
        data = bytearray(nbytes)
        cl.enqueue_copy(
            queue=self.context.queue,
            dest=data,  # nbytes taken from min(len(data),len(buffer))
            src=self.buffer,
            device_offset=offset,
        )
        return data

    def to_pointer_arg(self, offset, nbytes):
        """return data that can be used as argument in kernel

        Can fail if offset is not a multiple of self.alignment

        """
        return self.buffer[offset : offset + nbytes]


class KernelPyopencl(object):
    def __init__(
        self,
        function,
        description,
        context,
        wait_on_call=True,
    ):

        self.function = function
        self.description = description
        self.context = context
        self.wait_on_call = wait_on_call

    def to_function_arg(self, arg, value):
        if arg.pointer:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                if isinstance(value, cl.Buffer):
                    return value
                elif hasattr(value, "dtype"):  # nparray
                    assert isinstance(value, cla.Array)
                    return value.base_data[value.offset :]
                elif hasattr(value, "_shape"):  # xobject array
                    raise NotImplementedError
                else:
                    raise NotImplementedError
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} "
                    f"of kernel {self.description.pyname}"
                )
        else:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                return arg.atype(value)  # try to return a numpy scalar
            elif hasattr(arg.atype, "_size"):  # it is a compound xobject
                return value._buffer.buffer[value._offset :]
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} of kernel {self.description.pyname}"
                )

    @property
    def num_args(self):
        return len(self.description.args)

    def __call__(self, **kwargs):
        assert len(kwargs.keys()) == self.num_args
        arg_list = []
        for arg in self.description.args:
            vv = kwargs[arg.name]
            arg_list.append(self.to_function_arg(arg, vv))

        if isinstance(self.description.n_threads, str):
            n_threads = kwargs[self.description.n_threads]
        else:
            n_threads = self.description.n_threads

        event = self.function(
            self.context.queue, (n_threads,), None, *arg_list
        )

        if self.wait_on_call:
            event.wait()

        return event


class FFTPyopencl(object):
    def __init__(self, context, data, axes, wait_on_call=True):

        self.context = context
        self.axes = axes
        self.wait_on_call = wait_on_call

        assert len(data.shape) > max(axes)

        # Check internal dimensions are powers of two
        for ii in axes[:-1]:
            nn = data.shape[ii]
            frac_part, _ = np.modf(np.log(nn) / np.log(2))
            assert np.isclose(frac_part, 0), (
                "PyOpenCL FFT requires"
                " all dimensions apart from the last to be powers of two!"
            )

        import gpyfft

        self._fftobj = gpyfft.fft.FFT(
            context.context, context.queue, data, axes=axes
        )

    def transform(self, data):
        """The transform is done inplace"""

        (event,) = self._fftobj.enqueue_arrays(data)
        if self.wait_on_call:
            event.wait()
        return event

    def itransform(self, data):
        """The transform is done inplace"""

        (event,) = self._fftobj.enqueue_arrays(data, forward=False)
        if self.wait_on_call:
            event.wait()
        return event


if _enabled:
    available.append(ContextPyopencl)



import numpy as np

from .typeutils import get_a_buffer, Info
from .scalar import Int64
from .array import Array

"""
union typ1 typ2 ...

Initialization:
    1) Union(instance)
    1) Union((typename,...))


Data layout:
    - typeid
    - data
    - _is_static_type

Array instance:
    _itemtypes: list of item types
"""


class MetaUnion(type):
    def __new__(cls, name, bases, data):
        if "_itemtypes" in data:
            itemtypes = data["_itemtypes"]
            typeids = {}
            types = {}
            isize = itemtypes[0]
            for ii, itemtype in enumerate(itemtypes):
                name = itemtype.__name__
                typeids[name] = ii
                types[name] = itemtype
                if itemtype._size != isize:
                    isize = None

            data["_typeids"] = typeids
            data["_types"] = types
            data["_size"] = isize

        return type.__new__(cls, name, bases, data)

    def __getitem__(cls, shape):
        return Array.mk_arrayclass(cls, shape)


class Union(metaclass=MetaUnion):
    _types: list
    _size: None
    _typeids: dict
    _typenames: dict
    _itemtypes: list

    @classmethod
    def _get_type_index(cls, value):
        return cls._typeids[type(value).__name__]

    @classmethod
    def _get_type(cls, name):
        return cls._typenames[name]

    @classmethod
    def add_type(cls, itemtype):
        name = itemtype.__name__
        if name not in cls._typenames:
            cls._itemtypes.append(itemtype)
            cls._types[itemtype.__name__] = itemtype
            cls._typeids[itemtype.__name__] = cls._itemtypes.index(itemtype)
        else:
            raise ValueError(f"{itemtype} already in union")

    @classmethod
    def _inspect_args(cls, *args):
        if len(args) == 1:
            value = args[0]
            if type(value) in cls._itemtypes:
                size = value._get_size() + 8
                return Info(
                    size=size,
                    typeid=cls._get_type_index(value),
                    is_raw=True,
                    value=value,
                )
            elif type(value) is tuple:
                typename, value = value
                itemtype = cls._types[typename]
                typeid = cls._typeids[typename]
                iinfo = itemtype._inspect_args(value)
                return Info(
                    size=iinfo.size + 8,
                    typeid=typeid,
                    extra=iinfo,
                    is_raw=False,
                    value=value,
                )
            else:
                raise ValueError(f"{value} has the wrong type")
        else:
            raise ValueError(f"{value} has wrong number of arguments")

    @classmethod
    def _from_buffer(cls, buffer, offset):
        self = object.__new__(cls)
        self._buffer = buffer
        self._offset = offset
        return self

    @classmethod
    def _to_buffer(cls, buffer, offset, value, info=None):
        if info is None:
            info = cls._inspect_args(value)
        Int64._to_buffer(buffer, offset, info.typeid)
        coffset = offset + 8
        if info.is_raw:
            info.itemtype._buffer.copy_from(
                coffset, value._buffer, value._offset, info.size
            )
        else:
            info.itemtype._to_buffer(buffer, coffset, value, info=info.extra)

    def __init__(self, *args, _context=None, _buffer=None, _offset=None):
        info = self.__class__._inspect_args(*args)
        self._buffer, self._offset = get_a_buffer(
            info.size, _context, _buffer, _offset
        )

        self.__class__._to_buffer(self._buffer, self._offset, info.value, info)

        self._size = info._size
        self._itemtype = info._itemtype

    def get(self):
        return self._itemtype._from_buffer(self._buffer, self._offset + 8)

    def _get_size(self):
        if self.__class__._size is None:
            return Int64._from_buffer(self._buffer, self._offset)
        else:
            return self.__class__._size



import numpy as np


def _patch_pyopencl_array(cl, cla, ctx):
    prg = cl.Program(
        ctx,
        """
        __kernel void copy_array_fcont(
                     const int    fcont, // bool not accepted
                     const int    ndim,
                     const int    nelem,
            __global const int*   shape,
                     const int    itemsize,
            __global const char*  buffer_src,
            __global const int*   strides_src,
                     const int    offset_src,
            __global       char*  buffer_dest,
            __global const int*   strides_dest,
                     const int    offset_dest
                          )
        {
          int gid = get_global_id(0);
          int ibyte, idim, this_shape, slice_size, this_index, flat_index;
          int pos_src, pos_dest, this_stride_src, this_stride_dest;

          slice_size = nelem;
          flat_index = gid;
          pos_src = offset_src;
          pos_dest = offset_dest;
          for (idim=0; idim<ndim; idim++){
            if (fcont){
               this_shape = shape[ndim-idim-1];              // for f contiguous
               this_stride_src = strides_src[ndim-idim-1];   // for f contiguous
               this_stride_dest = strides_dest[ndim-idim-1]; // for f contiguous
               }
            else {
               this_shape = shape[idim];              // for c contiguous
               this_stride_src = strides_src[idim];   // for c contiguous
               this_stride_dest = strides_dest[idim]; // for c contiguous
            }


            slice_size = slice_size/this_shape;
            this_index = flat_index/slice_size;
            flat_index = flat_index - this_index*slice_size;

            pos_src = pos_src + this_index * this_stride_src;
            pos_dest = pos_dest + this_index * this_stride_dest;

          }

          for (ibyte=0; ibyte<itemsize; ibyte++){
            buffer_dest[pos_dest + ibyte] = buffer_src[pos_src + ibyte];
            }
        }
        """,
    ).build()

    knl_copy_array_fcont = prg.copy_array_fcont

    def _infer_fccont(arr):
        if arr.strides[0] < arr.strides[-1]:
            return "F"
        else:
            return "C"

    def copy_non_cont(src, dest, custom_itemsize=None, skip_typecheck=False):

        assert src.shape == dest.shape

        # The case float -> complex just works (by using the src itemsize)
        if not (src.dtype == np.float64 and dest.dtype == np.complex128):
            if not skip_typecheck:
                assert src.dtype == dest.dtype

        if src.strides[0] != src.strides[-1]:  # check is needed for 1d arrays
            assert _infer_fccont(src) == _infer_fccont(dest)

        if custom_itemsize is not None:
            itemsize = np.int32(custom_itemsize)
        else:
            itemsize = np.int32(src.dtype.itemsize)

        fcontiguous = 0
        if _infer_fccont(dest) == "F":
            fcontiguous = 1
        fcont = np.int32(fcontiguous)
        shape = cla.to_device(dest.queue, np.array(src.shape, dtype=np.int32))
        ndim = np.int32(len(shape))
        nelem = np.int32(np.prod(src.shape))
        buffer_src = src.base_data
        strides_src = cla.to_device(
            dest.queue, np.array(src.strides, dtype=np.int32)
        )
        offset_src = np.int32(src.offset)
        buffer_dest = dest.base_data
        strides_dest = cla.to_device(
            dest.queue, np.array(dest.strides, dtype=np.int32)
        )
        offset_dest = np.int32(dest.offset)

        event = knl_copy_array_fcont(
            dest.queue,
            (nelem,),
            None,
            # args:
            fcont,
            ndim,
            nelem,
            shape.data,
            itemsize,
            buffer_src,
            strides_src.data,
            offset_src,
            buffer_dest,
            strides_dest.data,
            offset_dest,
        )
        event.wait()

    def mysetitem(self, *args, **kwargs):
        try:
            self._old_setitem(*args, **kwargs)
        except (NotImplementedError, ValueError):
            dest = self[args[0]]
            src = args[1]
            if np.isscalar(src):
                src = dest._cont_zeros_like_me() + src
            copy_non_cont(src, dest)

    def mycopy(self):
        res = self._cont_zeros_like_me()
        copy_non_cont(self, res)
        return res

    def myreal(self):
        assert self.dtype == np.complex128
        res = cla.zeros(
            self.queue,
            shape=self.shape,
            dtype=np.float64,
            order=_infer_fccont(self),
        )
        copy_non_cont(self, res, custom_itemsize=8, skip_typecheck=True)
        return res

    def myget(self):
        try:
            return self._old_get()
        except AssertionError:
            return self.copy().get()

    def _cont_zeros_like_me(self):
        res = cla.zeros(
            self.queue,
            shape=self.shape,
            dtype=self.dtype,
            order=_infer_fccont(self),
        )
        return res

    # sum not implemented by pyopencl, I add it
    def mysum(self):
        dtype = getattr(np, self.dtype.name)
        try:
            res = dtype(cla.sum(self).get())
        except RuntimeError:
            res = dtype(cla.sum(self.copy()).get())

        return res

    cla.Array._cont_zeros_like_me = _cont_zeros_like_me

    if not hasattr(cla.Array, "_old_copy"):
        cla.Array._old_copy = cla.Array.copy
    cla.Array.copy = mycopy

    if not hasattr(cla.Array, "_old_setitem"):
        cla.Array._old_setitem = cla.Array.__setitem__
    cla.Array.__setitem__ = mysetitem

    if not hasattr(cla.Array, "_old_get"):
        cla.Array._old_get = cla.Array.get
    cla.Array.get = myget

    cla.Array.real = property(myreal)
    cla.Array.sum = mysum



import os


def specialize_source(source, specialize_for, search_in_folders=[]):

    assert specialize_for in ["cpu_serial", "cpu_openmp", "opencl", "cuda"]

    source_lines = source.splitlines()

    lines = []
    for ll in source_lines:
        if "//include_file" in ll:
            assert " for_context " in ll
            fname = (
                ll.split("//include_file")[-1].split("for_context")[0].strip()
            )
            temp_contexts = ll.split("for_context")[-1].split()
            temp_contexts = [ss.strip() for ss in temp_contexts]
            if specialize_for in temp_contexts:
                for fold in ["./"] + search_in_folders:
                    fpath = fold + "/" + fname
                    if os.path.isfile(fpath):
                        break
                else:
                    raise IOError(f"File {fname} not found")
                with open(fpath, "r") as fid:
                    flines = fid.readlines()
                lines.append("\n//from file: " + fname + "\n")
                for fll in flines:
                    lines.append(fll.rstrip())
                lines.append("\n//end file: " + fname + "\n")
        else:
            lines += ll.split("\n")

    indent = False
    new_lines = []
    inside_vect_block = False
    for ii, ll in enumerate(lines):
        if "//vectorize_over" in ll:
            if inside_vect_block:
                raise ValueError(f"Line {ii}: Previous vect block not closed!")
            inside_vect_block = True
            varname, limname = ll.split("//vectorize_over")[-1].split()
            if specialize_for.startswith("cpu"):
                new_lines.append(
                    f"for (int {varname}=0; {varname}<{limname}; {varname}++)"
                    + "{ //autovectorized\n"
                )
            elif specialize_for == "opencl":
                new_lines.append(f"int {varname}; //autovectorized\n")
                new_lines.append(
                    f"{varname}=get_global_id(0); //autovectorized\n"
                )

            elif specialize_for == "cuda":
                new_lines.append(f"int {varname}; //autovectorized\n")
                new_lines.append(
                    f"{varname}=blockDim.x * blockIdx.x + threadIdx.x;"
                    "//autovectorized\n"
                    f"if ({varname}<{limname})" + "{"
                )
        elif "//end_vectorize" in ll:
            if specialize_for.startswith("cpu"):
                new_lines.append("}//end autovectorized\n")
            elif specialize_for == "opencl":
                new_lines.append("//end autovectorized\n")
            elif specialize_for == "cuda":
                new_lines.append("}//end autovectorized\n")

            inside_vect_block = False
        else:
            if "//only_for_context" in ll:
                temp_contexts = ll.split("//only_for_context")[-1].split()
                temp_contexts = [ss.strip() for ss in temp_contexts]
                if specialize_for not in temp_contexts:
                    ll = "//" + ll
            if indent and inside_vect_block:
                new_lines.append("    " + ll)
            else:
                new_lines.append(ll)

    newfilecontent = "\n".join(new_lines)
    newfilecontent = newfilecontent.replace(
        "/*gpukern*/",
        {
            "cpu_serial": " ",
            "cpu_openmp": " ",
            "opencl": " __kernel ",
            "cuda": "__global__",
        }[specialize_for],
    )
    newfilecontent = newfilecontent.replace(
        "/*gpufun*/",
        {
            "cpu_serial": " static inline",
            "cpu_openmp": " static inline",
            "opencl": " ",
            "cuda": " __device__ ",
        }[specialize_for],
    )
    newfilecontent = newfilecontent.replace(
        "/*gpuglmem*/",
        {
            "cpu_serial": " ",
            "cpu_openmp": " ",
            "opencl": " __global ",
            "cuda": " ",
        }[specialize_for],
    )
    newfilecontent = newfilecontent.replace(
        "/*restrict*/",
        {
            "cpu_serial": " restrict ",
            "cpu_openmp": " restrict ",
            "opencl": "",
            "cuda": "",
        }[specialize_for],
    )

    return newfilecontent



from .context import Kernel, Arg

from .scalar import Int64, Void, Int8, is_scalar
from .struct import is_field, is_struct
from .array import is_index, is_array
from .ref import is_unionref, is_ref
from .string import is_string


def is_compound(atype):
    """Types that are referenced in C with an opaque pointer"""
    return is_struct(atype) or is_array(atype) or is_unionref(atype)


def is_type(atype):
    return is_compound(atype) or is_scalar(atype) or is_string(atype)


def get_inner_type(part):
    """type contained in a field, array or ref, else None"""
    if is_field(part):  # is a field
        return part.ftype
    elif is_array(part):  # is an array
        return part._itemtype
    elif is_ref(part):
        return part._reftype
    elif is_unionref(part):
        return None
    else:
        raise ValueError(f"Cannot get inner type of {part}")


def gen_pointer(ctype, conf):
    gpumem = conf.get("gpumem", "")
    return f"{gpumem}{ctype}"


def gen_c_type_from_arg(arg: Arg, conf):
    gpumem = conf.get("gpumem", "")
    if arg is None:
        cdec = "void"
    else:
        cdec = arg.atype._c_type
        if arg.pointer:
            cdec = f"{gpumem}{cdec}*"
        if arg.const:
            cdec = "const " + cdec
    return cdec


def gen_c_arg_from_arg(arg: Arg, conf):
    gpumem = conf.get("gpumem", "")
    cpurestrict = conf.get("cpurestrict", "")
    if arg is None:
        cdec = "void"
    else:
        cdec = arg.atype._c_type
        if arg.pointer:
            cdec = f"{gpumem}{cdec}*{cpurestrict}"
        elif is_compound(arg.atype):
            cdec = f"{cdec}{cpurestrict}"
        if arg.const:
            cdec = "const " + cdec
    return f"{cdec} {arg.name}"


def gen_c_size_from_arg(arg: Arg, conf):
    if arg is None:
        return None
    else:
        if is_compound(arg.atype):
            return conf.get("pointersize", 8)
        else:
            return arg.atype._size


def gen_c_decl_from_kernel(kernel: Kernel, conf):
    args = ", ".join([gen_c_arg_from_arg(arg, conf) for arg in kernel.args])
    if kernel.ret is None:
        ret = "void"
    else:
        ret = gen_c_type_from_arg(kernel.ret, conf)
    gpufun = conf.get("gpufun")
    if gpufun is None:
        return f"{ret} {kernel.c_name}({args})"
    else:
        return f"{gpufun} {ret} {kernel.c_name}({args})"


def get_layers(parts):
    layers = 0
    for part in parts:
        if hasattr(part, "_shape"):
            layers += 1
    return layers


def int_from_obj(offset, conf):
    inttype = gen_pointer(conf.get("inttype", "int64_t") + "*", conf)
    chartype = gen_pointer(conf.get("chartype", "char") + "*", conf)
    return f"*({inttype})(({chartype}) obj+{offset})"


def Field_get_c_offset(self, conf):
    if self.is_reference:
        doffset = f"offset+{self.offset}"  # starts of data
        refoffset = int_from_obj(doffset, conf)
        return [f"  offset+={refoffset};"]
    else:
        return self.offset


def Ref_get_c_offset(self, conf):
    refoffset = int_from_obj("offset", conf)
    return [f"  offset+={refoffset};"]


def Index_get_c_offset(part, conf, icount):
    cls = part.cls
    inttype = conf.get("inttype", "int64_t")

    out = []
    if hasattr(cls, "_strides"):  # static shape or 1d dynamic shape
        strides = cls._strides
    else:
        nd = len(cls._shape)
        strides = []
        stride_offset = 8 + nd * 8
        for ii in range(nd):
            sname = f"{inttype} {cls.__name__}_s{ii}"
            svalue = int_from_obj(f"offset+{stride_offset}", conf)
            out.append(f"{sname}={svalue};")
            strides.append(sname)

    soffset = "+".join([f"i{ii+icount}*{ss}" for ii, ss in enumerate(strides)])
    if cls._data_offset > 0:
        soffset = f"{cls._data_offset}+{soffset}"
    if cls._is_static_type:
        out.append(f"  offset+={soffset};")
    else:
        out.append(int_from_obj(f"offset+{soffset}", conf))
    return out


def gen_method_offset(path, conf):
    """return code to obtain offset of the target in bytes"""
    inttype = conf.get("inttype", "int64_t")
    lst = [f"  {inttype} offset=0;"]
    offset = 0
    icount = 0
    for part in path:
        if is_index(part):
            soffset = Index_get_c_offset(part, conf, icount)
            icount += len(part.cls._shape)
        elif is_field(part):
            soffset = Field_get_c_offset(part, conf)
        elif is_ref(part):
            soffset = Ref_get_c_offset(part, conf)
        else:
            soffset = None
        if type(soffset) is int:
            offset += soffset
        elif type(soffset) is list:
            if offset > 0:
                lst.append(f"  offset+={offset};")  # dump current offset
            lst.extend(soffset)  # update reference offset
            offset = 0
    if offset > 0:
        lst.append(f"  offset+={offset};")
    return "\n".join(lst)


def gen_fun_kernel(cls, path, action, const, extra, ret, add_nindex=False):
    typename = cls._c_type
    fields = []
    indices = 0
    for part in path:
        if is_field(part):  # is field
            fields.append(part.name)
        elif is_index(part):  # is array
            indices += len(part.cls._shape)
    if add_nindex and indices > 0:
        action += str(indices)
    fun_name = [typename, action]
    if len(fields) > 0:
        fun_name.append("_".join(fields))
    fun_name = "_".join(fun_name)
    args = [Arg(cls, pointer=False, const=const, name="obj")]
    for ii in range(indices):
        args.append(Arg(Int64, name=f"i{ii}"))

    args.extend(extra)

    return Kernel(args, c_name=fun_name, ret=ret)


def gen_c_pointed(target: Arg, conf):
    size = gen_c_size_from_arg(target, conf)
    ret = gen_c_type_from_arg(target, conf)
    if target.pointer or is_compound(target.atype):
        chartype = gen_pointer(conf.get("chartype", "char") + "*", conf)
        return f"({ret})(({chartype}) obj+offset)"
    else:
        rettype = gen_pointer(ret + "*", conf)
        if size == 1:
            return f"*(({rettype}) obj+offset)"
        else:
            chartype = gen_pointer(conf.get("chartype", "char") + "*", conf)
            return f"*({rettype})(({chartype}) obj+offset)"


def gen_method_get(cls, path, conf):
    lasttype = path[-1]
    retarg = Arg(lasttype)
    kernel = gen_fun_kernel(
        cls,
        path,
        const=True,
        action="get",
        extra=[],
        ret=retarg,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]
    lst.append(gen_method_offset(path, conf))
    pointed = gen_c_pointed(retarg, conf)
    lst.append(f"  return {pointed};")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_set(cls, path, conf):
    lasttype = path[-1]
    valarg = Arg(lasttype, name="value")
    kernel = gen_fun_kernel(
        cls,
        path,
        const=False,
        action="set",
        extra=[valarg],
        ret=None,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]
    lst.append(gen_method_offset(path, conf))
    pointed = gen_c_pointed(valarg, conf)
    lst.append(f"  {pointed}=value;")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_getp(cls, path, conf):
    lasttype = path[-1]
    if lasttype is None:
        retarg = Arg(Void, pointer="True")
    else:
        retarg = Arg(lasttype)
    if is_scalar(lasttype):
        retarg.pointer = True

    kernel = gen_fun_kernel(
        cls,
        path,
        const=False,
        action="getp",
        extra=[],
        ret=retarg,
        add_nindex=True,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]
    lst.append(gen_method_offset(path, conf))
    pointed = gen_c_pointed(retarg, conf)
    lst.append(f"  return {pointed};")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_len(cls, path, conf):
    lasttype = path[-1]

    retarg = Arg(Int64)

    kernel = gen_fun_kernel(
        cls,
        path,
        const=False,
        action="len",
        extra=[],
        ret=retarg,
        add_nindex=True,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]

    if lasttype._is_static_shape:
        ll = lasttype._get_n_items()
        lst.append(f"  return {ll};"), kernel
    else:
        lst.append(gen_method_offset(path, conf))
        arrarg = Arg(Int64, pointer=True)
        pointed = gen_c_pointed(arrarg, conf)
        typearr = gen_pointer("int64_t*", conf)
        lst.append(f"  {typearr} arr= {pointed};")
        terms = []
        ii = 1
        for sh in lasttype._shape:
            if sh is None:
                terms.append(f"arr[{ii}]")
            else:
                terms.append(str(sh))
        terms = "*".join(terms)
        lst.append(f"  return {terms};")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_size(cls, path, conf):
    innertype = get_inner_type(path[-1])
    if innertype is None:  # cannot determine size
        return (None, None)
    retarg = Arg(Int64)

    action = "size"
    layers = get_layers(path)
    if layers > 0 and not is_scalar(innertype):
        action += str(layers)

    kernel = gen_fun_kernel(
        cls,
        path,
        const=False,
        action=action,
        extra=[],
        ret=retarg,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]

    if innertype._size is None:
        lst.append(gen_method_offset(path, conf))
        pointed = gen_c_pointed(retarg, conf)
        lst.append(f"  return {pointed};")
    else:
        lst.append(f"  return {innertype._size};")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_shape(cls, path, conf):
    "return shape in an array"
    return None, None


def gen_method_nd(cls, path, conf):
    "return length of shape"
    return None, None


def gen_method_strides(cls, path, conf):
    "return strides"
    return None, None


def gen_method_getpos(cls, path, conf):
    "return slot position from index and strides"
    return None, None


def gen_method_typeid(cls, path, conf):
    "return typeid of a unionref"
    retarg = Arg(Int64)

    kernel = gen_fun_kernel(
        cls,
        path,
        const=True,
        action="typeid",
        extra=[],
        ret=retarg,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]

    lst.append(gen_method_offset(path, conf))
    pointed = gen_c_pointed(retarg, conf)
    lst.append("  offset+=8;")
    lst.append(f"  return {pointed};")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_member(cls, path, conf):
    "return typeid of a unionref"
    retarg = Arg(Void, pointer=True)

    kernel = gen_fun_kernel(
        cls,
        path,
        const=True,
        action="member",
        extra=[],
        ret=retarg,
    )
    decl = gen_c_decl_from_kernel(kernel, conf)

    lst = [decl + "{"]

    lst.append(gen_method_offset(path, conf))

    # pointed = gen_c_pointed(Arg(Int8, pointer=True), conf)
    lst.extend(Ref_get_c_offset("offset",conf))
    pointed = gen_c_pointed(retarg, conf)
    lst.append(f" return {pointed};")
    # GPU not handled
    #lst.append("  char *reloff_p = (char *) obj + offset;")
    #lst.append("  int64_t  reloff= *(int64_t *) reloff_p;")
    #lst.append("  return (void*) (reloff_p + reloff);")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_method_switch(cls, path, conf, method):
    "generate switch methods declared in _methods"
    lasttype = path[-1]

    kernel = gen_fun_kernel(
        cls,
        path,
        const=True,
        action=method.c_name,
        extra=method.args,
        ret=method.ret,
    )
    refname=lasttype.__name__

    decl = gen_c_decl_from_kernel(kernel, conf)
    lst = [decl + "{"]
    voidp=gen_pointer("void*",conf)
    lst.append(f"  {voidp} member = {refname}_member(obj);")
    lst.append(f"  switch ({refname}_typeid(obj)){{")
    for atype in lasttype._reftypes:
        atname=atype.__name__
        targs=[f"({atname}) member"]
        for arg in kernel.args[1:]:
            targs.append(f"{arg.name}")
        targs=','.join(targs)
        lst.append(f"""\
        #ifndef {refname.upper()}_SKIP_{atname.upper()}
        case {refname}_{atname}_t:
            return {atname}_{method.c_name}({targs});
            break;
        #endif"""
        )
    lst.append ("  }")
    lst.append ("  return 0;")
    lst.append("}")
    return "\n".join(lst), kernel


def gen_typedef(cls, conf):
    gpumem = conf.get("gpumem", "")
    typename = cls._c_type
    return f"typedef {gpumem} struct {typename}_s * {typename};"


def gen_enum(cls, conf):
    typename = cls.__name__
    lst = ",".join(f"{typename}_{tt._c_type}_t" for tt in cls._reftypes)
    return f"enum {typename}_e{{{lst}}};"


def methods_from_path(cls, path, conf):
    """
    size: all
    get, set: lasttype is scalar
    getp: all real types

    """
    out = []
    lasttype = path[-1]

    if is_scalar(lasttype):
        out.append(gen_method_get(cls, path, conf))
        out.append(gen_method_set(cls, path, conf))

    if is_type(lasttype):
        out.append(gen_method_getp(cls, path, conf))

    if is_array(lasttype):
        out.append(gen_method_len(cls, path, conf))
    #    out.append(gen_method_shape(cls, path, conf))
    #    out.append(gen_method_nd(cls, path, conf))
    #    out.append(gen_method_strides(cls, path, conf))
    #    out.append(gen_method_getpos(cls, path, conf))

    if is_unionref(lasttype):
        out.append(gen_method_typeid(cls, path, conf))
        out.append(gen_method_member(cls, path, conf))
        if cls == lasttype:
            for method in lasttype._methods:
                out.append(gen_method_switch(cls, path, conf, method))
    return out


def gen_cdef(cls, conf):
    out = []
    out.append(gen_typedef(cls, conf))
    if is_unionref(cls):
        out.append(gen_enum(cls, conf))
    return "\n".join(out)


def gen_code(cls, paths, conf):
    """
    Generate source code, used by add_kernels

    Provide all symbols starting from cls.__name__

    The definition must be completed by inner classes if present
    """

    typename = cls.__name__
    sources = []
    sources.append(f"#ifndef XOBJ_TYPEDEF_{typename}")
    sources.append(f"#define XOBJ_TYPEDEF_{typename}")
    sources.append(gen_cdef(cls, conf))

    out = []
    for path in paths:
        out.extend(methods_from_path(cls, path, conf))

    for source, _ in out:
        if source is not None:
            sources.append(source)

    sources.append(f"#endif")

    source = "\n".join(sources)

    return source


def gen_kernels(cls, paths, conf):
    """
    Generate kernel defintions of the C Api

    """
    out = []
    for path in paths:
        out.extend(methods_from_path(cls, path, conf))

    kernels = {}
    for _, kernel in out:
        if kernel is not None:
            kernels[kernel.c_name] = kernel

    return kernels


def gen_cdefs(cls, paths, conf):
    """

    Generate kernel defintions of the C Api

    """

    kernels = gen_kernels(cls, paths, conf)

    out = [gen_cdef(cls, conf)]

    for _, kernel in kernels.items():
        out.append(gen_c_decl_from_kernel(kernel, conf) + ";")

    return "\n".join(out)



"""


struct <name> field1 type1 ... fieldn typen

Layout:
  [ instance size ]: is not static
  static-field1
  ..
  static-fieldn
  [ offset field 2 ]
  [ offset ...
  [ offset field n ]
  [ dynamic-field1 ]
  [ ...
  [ dynamic-fieldn ]

Current implementation:
    1) instance stores the actual offsets for dynamic offset. Although it is wasteful for memory, it avoids a double round trip. This hinges on the structure being frozen at initializations.

Struct class:
- _size: class size, None if not static
- _fields: list of fields
- _d_fields: list of dynamic fields
- _s_fields: list of static fields

Struct instance:
- _offsets: cached offsets of dynamic fields dict indexed by field.index
- _sizes: cached sizes of dynamic fields dict indexed by field.index
- _size: cached size of the object

Field instance:
- ftype
- index
- name
- offset
- is_static_type
- has_update
- readonly


"""
import logging
from typing import Callable, Optional


from .typeutils import (
    get_a_buffer,
    dispatch_arg,
    Info,
    _to_slot_size,
    _is_dynamic,
    default_conf,
)

from .scalar import Int64
from .array import Array


log = logging.getLogger(__name__)


class Field:
    def __init__(
        self, ftype, default=None, readonly=False, default_factory=None
    ):
        self.ftype = ftype
        self.default = default
        self.default_factory = default_factory
        self.index = None  # filled by class creation
        self.name = None  # filled by class creation
        self.offset = None  # filled by class creation
        self.is_reference = None  # filled by class creation
        self.has_update = False  # filled by class creation
        self.readonly = readonly
        self.is_union = None  # filled by class creation

    def __repr__(self):
        return f"<field{self.index} {self.name} at {self.offset}>"

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        else:
            ftype, offset = self.get_offset(instance)
            return ftype._from_buffer(instance._buffer, offset)

    def __set__(self, instance, value):
        """
        A value can be:
          - python value: dict or scalar or list...
          - another instance of the same type
          - ???a buffer_protocol object???
        """
        if self.readonly:
            raise AttributeError(f"Field {self.name} is read-only")

        if hasattr(self.ftype, "_update"):
            self.__get__(instance)._update(value)
        else:  # TODO check if below is really needed
            ftype, offset = self.get_offset(instance)
            ftype._to_buffer(instance._buffer, offset, value)

    def get_offset(self, instance):  # compatible with info
        if self.is_reference:
            reloffset = instance._offsets[self.index]
            if self.is_union:
                absoffset = instance._offset + reloffset
                ftype = self.ftype._get_stored_type(
                    instance._buffer, absoffset
                )
            else:
                ftype = self.ftype
        else:
            reloffset = self.offset
            ftype = self.ftype
        return ftype, instance._offset + reloffset

    def get_default(self):
        if self.default_factory is None:
            if self.default is None:
                return self.ftype()
            else:
                return dispatch_arg(self.ftype, self.default)
        else:
            return self.default_factory()

    def value_from_args(self, arg):
        if self.name in arg:
            return arg[self.name]
        else:
            return self.get_default()


class MetaStruct(type):
    def __new__(cls, name, bases, data):
        offset = 0
        fields = []
        s_fields = []
        d_fields = []
        is_static = True
        findex = 0
        for aname, field in data.items():
            if hasattr(field, "_inspect_args"):
                data[aname] = Field(field)
        for aname, field in data.items():
            if isinstance(field, Field):
                field.index = findex
                findex += 1
                field.name = aname
                if hasattr(field.ftype, "_update"):
                    field.has_update = True
                fields.append(field)
                if field.ftype._size is None:
                    d_fields.append(field)
                    is_static = False
                else:
                    s_fields.append(field)
        data["_fields"] = fields
        data["_s_fields"] = s_fields
        data["_d_fields"] = d_fields

        if is_static:
            offset = 0
            for field in fields:
                field.offset = offset
                field.is_reference = False
                offset += _to_slot_size(field.ftype._size)
            size = offset

            def _get_size(self):
                return self.__class__._size

            def _inspect_args(cls, *args, **kwargs):
                info = Info(size=cls._size, is_static_size=True)
                if len(args) == 1:
                    info.value = args[0]
                else:
                    info.value = kwargs
                return info

        else:
            size = None
            for field in fields:
                offset = 8  # first slot is instance size
                for field in s_fields:
                    field.offset = offset
                    field.is_reference = False
                    offset += _to_slot_size(field.ftype._size)
                # other dynamic fields
                for field in d_fields[1:]:
                    field.offset = offset
                    field.is_reference = True
                    offset += _to_slot_size(8)
                # first dynamic field
                d_fields[0].offset = offset
                d_fields[0].is_reference = False

            def _get_size(self):
                return Int64._from_buffer(self._buffer, self._offset)

            def _inspect_args(cls, *args, **kwargs):
                log.debug(f"get size for {cls} from {args} {kwargs}")
                info = Info()
                if len(args) == 1:  # is a dict or xobj
                    arg = args[0]
                    info.value = arg
                    if isinstance(arg, dict):
                        offsets = {}  # offset of dynamic data
                        extra = {}
                        offset = d_fields[
                            0
                        ].offset  # offset first dynamic data
                        log.debug(f"{arg}")
                        for field in cls._d_fields:
                            farg = field.value_from_args(arg)
                            log.debug(
                                f"get size for field `{field.name}` using `{farg}`"
                            )
                            finfo = dispatch_arg(
                                field.ftype._inspect_args, farg
                            )
                            if hasattr(finfo, "_offsets"):  # is dinamic
                                extra[field.index] = finfo
                            offsets[field.index] = offset
                            offset += _to_slot_size(finfo.size)
                        # _offsets is used to because of field.get_offset(info)
                        info.size = offset
                        info._offsets = offsets
                        if len(extra) > 0:
                            info.extra = extra
                    elif isinstance(arg, cls):
                        info.size = arg._get_size()
                        info._offsets = {
                            kk: vv for kk, vv in arg._offsets.items()
                        }
                    else:
                        raise ValueError(f"{arg} Not valid type for {cls}")
                else:  # python argument
                    return cls._inspect_args(kwargs)
                return info

        data["_size"] = size
        data["_get_size"] = _get_size
        data["_inspect_args"] = classmethod(_inspect_args)
        if "_c_type" not in data:
            data["_c_type"] = name

        return type.__new__(cls, name, bases, data)

    def __getitem__(cls, shape):
        return Array.mk_arrayclass(cls, shape)

    def __repr__(cls):
        return f"<struct {cls.__name__}>"


class Struct(metaclass=MetaStruct):
    _fields: list
    _d_fields: list
    _inspect_args: Callable
    _size: Optional[int]

    @classmethod
    def _pre_init(cls, *args, **kwargs):
        return args, kwargs

    def _post_init(self):
        pass

    @classmethod
    def _from_buffer(cls, buffer, offset):
        self = object.__new__(cls)
        self._buffer = buffer
        self._offset = offset
        _offsets = {}
        for field in self._d_fields:
            offset = self._offset + field.offset
            val = Int64._from_buffer(self._buffer, offset)
            _offsets[field.index] = val
        self._offsets = _offsets
        self._size = self._get_size()
        self._post_init()
        return self

    @classmethod
    def _to_buffer(cls, buffer, offset, value, info=None):
        if isinstance(value, cls):  # binary copy
            buffer.update_from_xbuffer(
                offset, value._buffer, value._offset, value._size
            )
        else:  # value must be a dict, again potential disctructive
            if info is None:
                info = cls._inspect_args(value)
            if cls._size is None:
                Int64._to_buffer(buffer, offset, info.size)
            if hasattr(
                info, "_offsets"
            ):  # if it has a least two dynamic fields
                cls._set_offsets(buffer, offset, info._offsets)
            extra = getattr(info, "extra", {})
            for field in cls._fields:
                fvalue = field.value_from_args(value)
                if field.is_reference:
                    foffset = offset + info._offsets[field.index]
                else:
                    foffset = offset + field.offset
                finfo = extra.get(field.index)
                field.ftype._to_buffer(buffer, foffset, fvalue, finfo)

    def _update(self, value):
        # check if direct copy is possible
        if isinstance(value, self.__class__) and value._size == self._size:
            self._buffer.update_from_xbuffer(
                self._offset, value._buffer, value._offset, value._size
            )
        else:
            for field in self._fields:
                if field.name in value:
                    field.__set__(self, value[field.name])

    def __init__(
        self, *args, _context=None, _buffer=None, _offset=None, **kwargs
    ):
        """
        Create new struct in buffer from offset.
        If offset not provide
        """
        cls = self.__class__
        args, kwargs = cls._pre_init(*args, **kwargs)
        # compute resources
        info = cls._inspect_args(*args, **kwargs)
        self._size = info.size
        # acquire buffer
        self._buffer, self._offset = get_a_buffer(
            info.size, _context, _buffer, _offset
        )
        # if dynamic struct store dynamic offsets
        if hasattr(info, "_offsets"):
            self._offsets = info._offsets  # struct offsets
        cls._to_buffer(self._buffer, self._offset, info.value, info)
        self._post_init()

    @classmethod
    def _set_offsets(cls, buffer, offset, loffsets):
        log.debug(f"{cls} set offset {loffsets}")
        for index, data_offset in loffsets.items():
            foffset = offset + cls._fields[index].offset
            Int64._to_buffer(buffer, foffset, data_offset)

    def _to_dict(self):
        return {field.name: field.__get__(self) for field in self._fields}

    def __iter__(self):
        for field in self._fields:
            yield field.name

    def __getitem__(self, key):
        for field in self._fields:
            if field.name == key:
                return field.__get__(self)
        raise KeyError("{key} not found")

    def __contains__(self, key):
        for field in self._fields:
            if field.name == key:
                return True
        else:
            return False

    def __repr__(self):
        fields = (
            (field.name, repr(getattr(self, field.name)))
            for field in self._fields
        )
        fields = ", ".join(f"{k}={v}" for k, v in fields)
        longform = f"{self.__class__.__name__}({fields})"
        if len(longform) > 60:
            return f"{self.__class__.__name__}(...)"
        else:
            return longform

    @classmethod
    def _gen_data_paths(cls, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [cls])
        for field in cls._fields:
            path = base + [cls, field]
            paths.append(path)
            if hasattr(field.ftype, "_gen_data_paths"):
                paths.extend(field.ftype._gen_data_paths(path))
        return paths

    @classmethod
    def _gen_c_api(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_code(cls, paths, conf)

    @classmethod
    def _gen_c_decl(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_cdefs(cls, paths, conf)

    @classmethod
    def _gen_kernels(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_kernels(cls, paths, conf)

    def _get_offset(self, fieldname):
        for ff in self._fields:
            if ff.name == fieldname:
                return ff.get_offset(self)[1]

    @classmethod
    def _get_inner_types(cls):
        return [fl.ftype for fl in cls._fields]


def is_struct(atype):
    return isinstance(atype, MetaStruct)


def is_field(atype):
    return isinstance(atype, Field)



import os
import uuid
import importlib
import sysconfig
import logging

import numpy as np

from .context import (
    XBuffer,
    XContext,
    ModuleNotAvailable,
    available,
    classes_from_kernels,
    sort_classes,
    sources_from_classes,
    _concatenate_sources,
)
from .specialize_source import specialize_source
from .linkedarray import BaseLinkedArray

log = logging.getLogger(__name__)

try:
    import cffi

    _enabled = True
except ImportError:
    log.info("cffi is not installed, ContextCPU will not be available")

    cffi = ModuleNotAvailable(
        message=("cffi is not installed. " "ContextCPU is not available!")
    )
    _enabled = False

try:
    import pyfftw

    pyfftw_available = True
except ImportError:
    log.info("pyfftw not available, will use numpy")
    pyfftw = ModuleNotAvailable(message="pyfftw not available")
    pyfftw_available = False

dtype_dict = {
    "float64": "double",
    "float32": "float",
    "int64": "int64_t",
    "int32": "int32_t",
    "int8": "int8_t",
    "uint64": "uint64_t",
    "uint32": "uint32_t",
}


def dtype2ctype(dtype):
    return dtype_dict[dtype.name]


def nplike_to_numpy(arr):
    return np.array(arr)


def cdef_from_kernel(kernel, pyname=None):
    if kernel.c_name is None:
        kernel.c_name = pyname
    if kernel.ret is not None:
        rettype = kernel.ret.get_c_type()
    else:
        rettype = "void"
    signature = f"{rettype} {kernel.c_name}("
    signature += ",".join(arg.get_c_type() for arg in kernel.args)
    signature += ");"
    return signature

if _enabled:
    # order of base classes matters as it defines which __setitem__ is used
    class LinkedArrayCpu(BaseLinkedArray, np.ndarray):

        @classmethod
        def _build_view(cls, a):
            assert len(a.shape) == 1
            return cls(shape=a.shape, dtype=a.dtype, buffer=a.data, offset=0,
                    strides=a.strides, order='C')

class ContextCpu(XContext):
    """

    Creates a CPU Platform object, that allows performing the computations
    on conventional CPUs.

    Returns:
         ContextCpu: platform object.

    """

    @property
    def nplike_array_type(self):
        return np.ndarray

    @property
    def linked_array_type(self):
        return LinkedArrayCpu

    def __init__(self, omp_num_threads=0):
        super().__init__()
        self.omp_num_threads = omp_num_threads

    def _make_buffer(self, capacity):
        return BufferNumpy(capacity=capacity, context=self)

    def add_kernels(
        self,
        sources=[],
        kernels=[],
        specialize=True,
        save_source_as=None,
        extra_compile_args=["-O3", "-Wno-unused-function"],
        extra_link_args=["-O3"],
        extra_cdef=None,
        extra_classes=[],
        extra_headers=[],
    ):

        """
        Adds user-defined kernels to to the context. The kernel source
        code is provided as a string and/or in source files and must contain
        the kernel names defined in the kernel descriptions.
        Args:
            sources (list): List of source codes that are concatenated before
                compilation. The list can contain strings (raw source code),
                File objects and Path objects.
            kernels (dict): Dictionary with the kernel descriptions
                in the form given by the following examples. The descriptions
                define the kernel names, the type and name of the arguments
                and identify one input argument that defines the number of
                threads to be launched (only on cuda/opencl).
            specialize_code (bool): If True, the code is specialized using
                annotations in the source code. Default is ``True``
            save_source_as (str): Filename for saving the specialized source
                code. Default is ```None```.
        Example:

        .. code-block:: python

            # A simple kernel
            src_code = '''
            /*gpukern*/
            void my_mul(const int n,
                /*gpuglmem*/ const double* x1,
                /*gpuglmem*/ const double* x2,
                /*gpuglmem*/       double* y) {
                int tid = 0 //vectorize_over tid
                y[tid] = x1[tid] * x2[tid];
                //end_vectorize
                }
            '''

            # Prepare description
            kernel_descriptions = {
                "my_mul": xo.Kernel(
                    args=[
                        xo.Arg(xo.Int32, name="n"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x1"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x2"),
                        xo.Arg(xo.Float64, pointer=True, const=False, name="y"),
                    ],
                    n_threads="n",
                    ),
            }

            # Import kernel in context
            ctx.add_kernels(
                sources=[src_code],
                kernels=kernel_descriptions,
                save_source_as=None,
            )

            # With a1, a2, b being arrays on the context, the kernel
            # can be called as follows:
            ctx.kernels.my_mul(n=len(a1), x1=a1, x2=a2, y=b)
        """

        classes = classes_from_kernels(kernels)
        classes.update(extra_classes)
        classes = sort_classes(classes)
        cls_sources = sources_from_classes(classes)

        headers = ["#include <stdint.h>"]

        if self.omp_num_threads > 0:
            headers = ["#include <omp.h>"] + headers

        headers += extra_headers

        sources = headers + cls_sources + sources

        source, folders = _concatenate_sources(sources)

        if specialize:
            if self.omp_num_threads > 0:
                specialize_for = "cpu_openmp"
            else:
                specialize_for = "cpu_serial"
            # included files are searched in the same folders od the src_filed
            specialized_source = specialize_source(
                source,
                specialize_for=specialize_for,
                search_in_folders=list(folders),
            )
        else:
            specialized_source = source

        if save_source_as is not None:
            with open(save_source_as, "w") as fid:
                fid.write(specialized_source)

        ffi_interface = cffi.FFI()

        cdefs = "\n".join(cls._gen_c_decl({}) for cls in classes)

        ffi_interface.cdef(cdefs)

        if extra_cdef is not None:
            ffi_interface.cdef(extra_cdef)

        for pyname, kernel in kernels.items():
            if pyname not in cdefs:  # check if kernel not already declared
                signature = cdef_from_kernel(kernel, pyname)
                ffi_interface.cdef(signature)
                log.debug(f"cffi def {pyname} {signature}")

        if self.omp_num_threads > 0:
            ffi_interface.cdef("void omp_set_num_threads(int);")

        # Generate temp fname
        tempfname = str(uuid.uuid4().hex)

        # Compile
        xtr_compile_args = ["-std=c99"]
        xtr_link_args = ["-std=c99"]
        xtr_compile_args += extra_compile_args
        xtr_link_args += extra_link_args
        if self.omp_num_threads > 0:
            xtr_compile_args.append("-fopenmp")
            xtr_link_args.append("-fopenmp")

        ffi_interface.set_source(
            tempfname,
            specialized_source,
            extra_compile_args=xtr_compile_args,
            extra_link_args=xtr_link_args,
        )

        ffi_interface.compile(verbose=True)

        # build full so filename, something like:
        # 0e14651ea79740119c6e6c24754f935e.cpython-38-x86_64-linux-gnu.so
        suffix = sysconfig.get_config_var("EXT_SUFFIX")
        so_fname = tempfname + suffix

        try:
            # Import the compiled module
            spec = importlib.util.spec_from_file_location(
                tempfname, os.path.abspath("./" + tempfname + suffix)
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            if self.omp_num_threads > 0:
                self.omp_set_num_threads = module.lib.omp_set_num_threads

            # Get the methods
            for pyname, kernel in kernels.items():
                self.kernels[pyname] = KernelCpu(
                    function=getattr(module.lib, kernel.c_name),
                    description=kernel,
                    ffi_interface=module.ffi,
                    context=self,
                )
                self.kernels[pyname].source = source
                self.kernels[pyname].specialized_source = specialized_source

        finally:
            # Clean temp files
            files_to_remove = [so_fname, tempfname + ".c", tempfname + ".o"]
            for ff in files_to_remove:
                if os.path.exists(ff):
                    os.remove(ff)

    def nparray_to_context_array(self, arr):
        """
        Moves a numpy array to the device memory. No action is performed by
        this function in the CPU context. The method is provided
        so that the CPU context has an identical API to the GPU ones.

        Args:
            arr (numpy.ndarray): Array to be transferred

        Returns:
            numpy.ndarray: The same array (no copy!).

        """
        return arr

    def nparray_from_context_array(self, dev_arr):
        """
        Moves an array to the device to a numpy array. No action is performed by
        this function in the CPU context. The method is provided so that the CPU
        context has an identical API to the GPU ones.

        Args:
            dev_arr (numpy.ndarray): Array to be transferred
        Returns:
            numpy.ndarray: The same array (no copy!)

        """
        return dev_arr

    @property
    def nplike_lib(self):
        """
        Module containing all the numpy features. Numpy members should be accessed
        through ``nplike_lib`` to keep compatibility with the other contexts.

        """

        return np

    def synchronize(self):
        """
        Ensures that all computations submitted to the context are completed.
        No action is performed by this function in the CPU context. The method
        is provided so that the CPU context has an identical API to the GPU ones.
        """
        pass

    def zeros(self, *args, **kwargs):
        """
        Allocates an array of zeros on the device. The function has the same
        interface of numpy.zeros"""
        return self.nplike_lib.zeros(*args, **kwargs)

    def plan_FFT(self, data, axes):
        """
        Generate an FFT plan object to be executed on the context.

        Args:
            data (numpy.ndarray): Array having type and shape for which the FFT
                needs to be planned.
            axes (sequence of ints): Axes along which the FFT needs to be
                performed.
        Returns:
            FFTCpu: FFT plan for the required array shape, type and axes.

        Example:

        .. code-block:: python

            plan = context.plan_FFT(data, axes=(0,1))

            data2 = 2*data

            # Forward tranform (in place)
            plan.transform(data2)

            # Inverse tranform (in place)
            plan.itransform(data2)
        """
        return FFTCpu(data, axes, threads=self.omp_num_threads)

    @property
    def kernels(self):

        """
        Dictionary containing all the kernels that have been imported to the context.
        The syntax ``context.kernels.mykernel`` can also be used.
        """

        return self._kernels


class BufferByteArray(XBuffer):
    def _make_context(self):
        return ContextCpu()

    def _new_buffer(self, capacity):
        return bytearray(capacity)

    def update_from_native(self, offset, source, source_offset, nbytes):
        """Copy data from native buffer into self.buffer starting from offset"""
        self.buffer[offset : offset + nbytes] = source[
            source_offset : source_offset + nbytes
        ]

    def copy_native(self, offset, nbytes):
        """return native data with content at from offset and nbytes"""
        return self.buffer[offset : offset + nbytes].copy()

    def copy_to_native(self, dest, dest_offset, source_offset, nbytes):
        """copy data from self.buffer into dest"""
        dest[dest_offset : dest_offset + nbytes] = self.buffer[
            source_offset : source_offset + nbytes
        ]

    def update_from_buffer(self, offset, source):
        """Copy data from python buffer such as bytearray, bytes, memoryview, numpy array.data"""
        nbytes = len(source)
        self.buffer[offset : offset + nbytes] = source

    def to_nplike(self, offset, dtype, shape):
        """view in nplike"""
        count = np.prod(shape)
        return np.frombuffer(
            self.buffer, dtype=dtype, count=count, offset=offset
        ).reshape(*shape)

    def update_from_nplike(self, offset, dest_dtype, value):
        value = nplike_to_numpy(value)
        if dest_dtype != value.dtype:
            value = value.astype(dtype=dest_dtype)
        self.update_from_native(offset, value.data, 0, value.nbytes)

    def to_bytearray(self, offset, nbytes):
        """copy in byte array: used in update_from_xbuffer"""
        return self.buffer[offset : offset + nbytes]

    def to_pointer_arg(self, offset, nbytes):
        """return data that can be used as argument in kernel"""
        return self.buffer[offset : offset + nbytes]


class BufferNumpy(XBuffer):
    def _make_context(self):
        return ContextCpu()

    def _new_buffer(self, capacity):
        return np.zeros(capacity, dtype="int8")

    def update_from_native(self, offset, source, source_offset, nbytes):
        """Copy data from native buffer into self.buffer starting from offset"""
        self.buffer[offset : offset + nbytes] = source[
            source_offset : source_offset + nbytes
        ]

    def copy_native(self, offset, nbytes):
        """return native data with content at from offset and nbytes"""
        return self.buffer[offset : offset + nbytes].copy()

    def copy_to_native(self, dest, dest_offset, source_offset, nbytes):
        """copy data from self.buffer into dest"""
        dest[dest_offset : dest_offset + nbytes] = self.buffer[
            source_offset : source_offset + nbytes
        ]

    def update_from_buffer(self, offset, source):
        """Copy data from python buffer such as bytearray, bytes, memoryview, numpy array.data"""
        nbytes = len(source)
        self.buffer[offset : offset + nbytes] = bytearray(source)

    def to_nplike(self, offset, dtype, shape):
        """view in nplike"""
        count = np.prod(shape)
        # dtype=np.dtype(dtype)
        # return self.buffer[offset:].view(dtype)[:count].reshape(*shape)
        return np.frombuffer(
            self.buffer, dtype=dtype, count=count, offset=offset
        ).reshape(*shape)

    def update_from_nplike(self, offset, dest_dtype, value):
        value = nplike_to_numpy(value)
        if dest_dtype != value.dtype:
            value = value.astype(dtype=dest_dtype)  # make a copy
        src = value.view("int8")
        self.buffer[offset : offset + src.nbytes] = value.view("int8")

    def to_bytearray(self, offset, nbytes):
        """copy in byte array: used in update_from_xbuffer"""
        return bytearray(self.buffer[offset : offset + nbytes])

    def to_pointer_arg(self, offset, nbytes):
        """return data that can be used as argument in kernel"""
        return self.buffer[offset : offset + nbytes]


class KernelCpu:
    def __init__(
        self,
        function,
        description,
        ffi_interface,
        context,
    ):

        self.function = function
        self.description = description
        self.ffi_interface = ffi_interface
        self.context = context

    def to_function_arg(self, arg, value):
        if arg.pointer:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                if hasattr(value, "dtype"):  # nparray
                    slice_first_elem = value[tuple(value.ndim * [slice(0, 1)])]
                    return self.ffi_interface.cast(
                        dtype2ctype(value.dtype) + "*",
                        self.ffi_interface.from_buffer(slice_first_elem.data),
                    )
                elif hasattr(value, "_shape"):  # xobject array
                    # TODO: check context
                    return self.ffi_interface.cast(
                        value._c_type + "*",
                        self.ffi_interface.from_buffer(
                            value._buffer.buffer[
                                value._offset + value._data_offset :
                            ]  # fails for pyopencl, cuda
                        ),
                    )
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} of kernel {self.description.pyname}"
                )
        else:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                return arg.atype(value)  # try to return a numpy scalar
            elif hasattr(arg.atype, "_size"):  # it is a compound xobject
                # TODO: check context
                buf = np.frombuffer(value._buffer.buffer, dtype="int8")
                ptr = buf.ctypes.data + value._offset
                return self.ffi_interface.cast(arg.atype._c_type, ptr)
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} of kernel {self.description.pyname}"
                )

    def from_function_arg(self, arg, value):
        return value

    @property
    def num_args(self):
        return len(self.description.args)

    def __call__(self, **kwargs):
        assert len(kwargs.keys()) == self.num_args
        arg_list = []
        for arg in self.description.args:
            vv = kwargs[arg.name]
            arg_list.append(self.to_function_arg(arg, vv))

        if self.context.omp_num_threads > 0:
            self.context.omp_set_num_threads(self.context.omp_num_threads)

        ret = self.function(*arg_list)

        if self.description.ret is not None:
            return self.from_function_arg(self.description.ret, ret)


class FFTCpu(object):
    def __init__(self, data, axes, threads=0):

        self.axes = axes
        self.threads = threads

        self.use_pyfftw = False
        if threads > 0 and pyfftw_available:
            self.use_pyfftw = True
            self.data = data
            self.data_temp = pyfftw.byte_align(0 * data)
            self.fftw = pyfftw.FFTW(
                data,
                self.data_temp,
                axes=axes,
                threads=threads,
                direction="FFTW_FORWARD",
                flags=("FFTW_MEASURE",),
            )
            self.ifftw = pyfftw.FFTW(
                data,
                self.data_temp,
                axes=axes,
                threads=threads,
                direction="FFTW_BACKWARD",
                flags=("FFTW_MEASURE",),
            )
            print(f"fftw simd_aligned={self.fftw.simd_aligned}")
            print(f"ifftw simd_aligned={self.fftw.simd_aligned}")
        else:
            # I perform one fft to have numpy cache the plan
            _ = np.fft.ifftn(np.fft.fftn(data, axes=axes), axes=axes)

    def transform(self, data):
        """The transform is done inplace"""
        if self.use_pyfftw:
            assert data is self.data
            self.fftw.execute()
            data[:] = self.data_temp[:]
        else:
            data[:] = np.fft.fftn(data, axes=self.axes)[:]

    def itransform(self, data):
        """The transform is done inplace"""
        if self.use_pyfftw:
            assert data is self.data
            self.ifftw.execute()
            data[:] = self.data_temp[:] / self.ifftw.N
        else:
            data[:] = np.fft.ifftn(data, axes=self.axes)[:]


if _enabled:
    available.append(ContextCpu)



"""
String: holds a null terminated variable length string

init:
    String(10)  create an empty string with  10 byte capacity (including termination \0)
    String("a string") create and initialize string with the minimal capacity

Layout:
    [capacity]
    data
    [null] at least 1 byte


String class data:
    _size = None
    _c_type = "char*"

String instance data:
    _size = string type size (including capacity)


TODO:
- String._to_buffer to use same context copy
- consider caching  the length
- consider using __slots__
- consider adding size in the class
"""

from .typeutils import get_a_buffer, Info, _to_slot_size, is_integer

from .scalar import Int64
from .array import Array

import logging

log = logging.getLogger(__name__)


class MetaString(type):
    def __getitem__(self, shape):
        return Array.mk_arrayclass(self, shape)

    def _inspect_args(cls, string_or_int):
        if cls._size is None:
            if isinstance(string_or_int, int):
                return Info(size=string_or_int + 8)
            elif isinstance(string_or_int, str):
                data = bytes(string_or_int, "utf8")
                size = _to_slot_size(len(data) + 1 + 8)
                return Info(size=size, data=data)  # add zero termination
            elif isinstance(string_or_int, cls):
                return Info(size=string_or_int._size)
            raise ValueError(
                f"String can accept only one integer or string and not `{string_or_int}`"
            )
        else:
            return Info(size=cls._size)

    def _to_buffer(cls, buffer, offset, value, info=None):
        log.debug(f"{cls} to buffer {offset}  `{value}`")
        if info is None:
            info = cls._inspect_args(value)
        size = info.size
        string_capacity = info.size - 8
        Int64._to_buffer(buffer, offset, size)
        if isinstance(value, String):
            buffer.update_from_xbuffer(
                offset, value._buffer, value._offset, value._size
            )
        elif isinstance(value, str):
            data = info.data
            off = string_capacity - len(data)
            data += b"\x00" * off
            log.debug(f"to_buffer {offset+8} {len(data)} {string_capacity}")
            buffer.update_from_buffer(offset + 8, data)
        elif is_integer(value):
            pass
        else:
            raise ValueError(f"{value} not a string")

    def _get_data(cls, buffer, offset):
        ll = Int64._from_buffer(buffer, offset)
        return buffer.to_bytearray(offset + 8, ll - 8)

    def _from_buffer(cls, buffer, offset):
        # TODO keep in mind that in windows many funcitons returns wchar encoded in utf16
        return cls._get_data(buffer, offset).decode("utf8").rstrip("\x00")

    def fixed(cls, size):
        if is_integer(size) and size > 0:
            name = f"String{size}"
            data = dict(_size=size)
            return MetaString(name, (String,), data)
        else:
            raise ValueError("String needs a positive integer")


class String(metaclass=MetaString):
    _size = None
    _c_type = "char*"

    def __init__(
        self, string_or_int, _buffer=None, _offset=None, _context=None
    ):
        info = self.__class__._inspect_args(string_or_int)
        size = info.size
        self._buffer, self._offset = get_a_buffer(
            size, _context, _buffer, _offset
        )

        self.__class__._to_buffer(
            self._buffer, self._offset, string_or_int, info=info
        )

        self._size = size

    def update(self, value):
        buffer = self._buffer
        offset = self._offset
        size = self._size
        if isinstance(value, String):
            if value._size < self._size:
                buffer.write(offset + 8, value.to_bytes())  # TODO use copy
            else:
                raise ValueError(f"{value} to large to fit in {size}")
        elif isinstance(value, str):
            info = self.__class__._inspect_args(value)
            if info > self._size:
                raise ValueError(f"{value} to large to fit in {size}")
            else:
                data = info.data
                off = _to_slot_size(size) - len(data)
                data += b"\x00" * off
                buffer.write(offset + 8, data)
        else:
            raise ValueError(f"{value} not a string")

    def to_str(self):
        return self.__class__._from_buffer(self._buffer, self._offset)

    def to_bytes(self):
        return self.__class__._get_data(self._buffer, self._offset)

    @classmethod
    def _gen_data_paths(cls, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [cls])
        return paths


def is_string(atype):
    return atype == String



from collections import defaultdict
from typing import NamedTuple, List, Type
from abc import ABC, abstractmethod
from pathlib import Path
import logging
import weakref
import os

"""

TODO:
    - add test get_a_buffer
    - Consider exposing Buffer and removing CLBuffer, ByteArrayBuffers..
    - Consider Buffer[offset] to create View and avoid _offset in type API
"""


log = logging.getLogger(__name__)


def topological_sort(source):
    """Sort tasks defined as {child: parents} in a list"""

    # Store parent count and parent dictionary.
    graph = {}
    num_parents = defaultdict(int)
    for child, parents in source.items():
        for parent in parents:
            graph.setdefault(parent, []).append(child)
            num_parents[child] += 1

    # Begin with the parent-less items.
    result = [child for child, parents in source.items() if len(parents) == 0]
    result.extend([item for item in graph if num_parents[item] == 0])

    # Descend through graph, removing parents as we go.
    for parent in result:
        if parent in graph:
            for child in graph[parent]:
                num_parents[child] -= 1
                if num_parents[child] == 0:
                    result.append(child)
            del graph[parent]

    # If there's a cycle, just throw in whatever is left over.
    has_cycle = bool(graph)
    if has_cycle:
        result.extend(list(graph.keys()))
    return result, has_cycle


def sort_classes(classes):
    cdict = {cls.__name__: cls for cls in classes}
    deps = {}
    lst = list(classes)
    for cls in lst:
        cldeps = []
        if hasattr(cls, "_get_inner_types"):
            for cl in cls._get_inner_types():
                if not cl.__name__ in cdict:
                    cdict[cl.__name__] = cl
                    lst.append(cl)
                cldeps.append(cl.__name__)
        if hasattr(cls, "_depends_on"):
            cldeps.extend(cls._depends_on)
        deps[cls.__name__] = cldeps
    lst, has_cycle = topological_sort(deps)
    if has_cycle:
        raise ValueError("Class dependencies have cycles")
    return [cdict[cn] for cn in lst if hasattr(cdict[cn], "_gen_c_api")]


def sources_from_classes(classes):
    sources = []
    for cls in classes:
        sources.append(cls._gen_c_api())
        if hasattr(cls, "_extra_c_source"):
            sources.append(cls._extra_c_source)
    return sources


def classes_from_kernels(kernels):
    classes = set()
    for _, kernel in kernels.items():
        classes.update(kernel.get_classes())
    return classes


def _concatenate_sources(sources):
    source = []
    folders = set()
    for ss in sources:
        if hasattr(ss, "read"):
            source.append(ss.read())
            folders.add(os.path.dirname(ss.name))
        elif isinstance(ss, Path):
            with open(ss, "r") as fid:
                source.append(fid.read())
            folders.add(ss.parent)
        else:
            source.append(ss)
    source = "\n".join(source)

    folders = [str(ff) for ff in folders]

    return source, folders


def _align(offset, alignment):
    "round to nearest multiple of 8"
    return (offset + alignment - 1) & (-alignment)


class MinimalDotDict(dict):
    def __getattr__(self, attr):
        if attr not in self:
            raise AttributeError(f"`{attr}` not found in dict")
        return self.get(attr)

    def __dir__(self):
        return list(self.keys())


class ModuleNotAvailable(object):
    def __init__(self, message="Module not available"):
        self.message = message

    def __getattr__(self, attr):
        raise NameError(self.message)


class XContext(ABC):
    minimum_alignment = 1

    def __init__(self):
        self._kernels = MinimalDotDict()
        self._buffers = []

    def new_buffer(self, capacity=1048576):
        buf = self._make_buffer(capacity=capacity)
        self.buffers.append(weakref.finalize(buf, log.debug, "free buf"))
        return buf

    @property
    def buffers(self):
        return self._buffers

    @property
    def kernels(self):
        return self._kernels

    @abstractmethod
    def _make_buffer(self, capacity):
        "return buffer"

    @abstractmethod
    def add_kernels(
        self,
        sources: list,
        kernels: dict,
        specialize: bool,
        save_source_as: str,
    ):
        pass

    @abstractmethod
    def nparray_to_context_array(self, arr):
        return arr

    @abstractmethod
    def nparray_from_context_array(self, dev_arr):
        return dev_arr

    @property
    @abstractmethod
    def nplike_lib(self):
        "return lib"

    @abstractmethod
    def synchronize(self):
        pass

    @abstractmethod
    def zeros(self, *args, **kwargs):
        "return arr"

    @abstractmethod
    def plan_FFT(self, data, axes):
        "return fft"


class XBuffer(ABC):
    def __init__(
        self,
        capacity=1048576,
        context=None,
        default_alignment=None,
        grow_step=None,
    ):

        if context is None:
            self.context = self._make_context()
        else:
            self.context = context
        self.buffer = self._new_buffer(capacity)
        self.capacity = capacity
        if default_alignment is None:
            default_alignment = self.context.minimum_alignment
        self.default_alignment = default_alignment
        self.chunks = [Chunk(0, capacity)]
        self.grow_step = grow_step

    @abstractmethod
    def _make_context(self):
        "return a default context"

    def allocate(self, size, align=True):
        # find available free slot
        # and update free slot if exists
        if align:
            alignment = self.default_alignment
        else:
            alignment = 1
        # sizepa = size + alignment - 1
        for chunk in self.chunks:
            offset = _align(chunk.start, alignment)
            newend = offset + size
            if chunk.end >= newend:
                chunk.start = newend
                if chunk.size == 0:
                    self.chunks.remove(chunk)
                return offset

        # no free slot check if can be allocated then try to grow
        sizepa = size + alignment - 1
        if sizepa > self.capacity:
            self.grow(sizepa)
        elif self.grow_step is not None:
            self.grow(self.grow_step)
        else:
            self.grow(self.capacity)

        # try again
        return self.allocate(size, align=align)

    def grow(self, capacity):
        """
        Add capacity to buffer
        """
        oldcapacity = self.capacity
        newcapacity = self.capacity + capacity
        newbuff = self._new_buffer(newcapacity)
        self.copy_to_native(
            dest=newbuff, dest_offset=0, source_offset=0, nbytes=oldcapacity
        )
        self.buffer = newbuff
        if len(self.chunks) == 0 or self.chunks[-1].end != self.capacity:
            self.chunks.append(Chunk(oldcapacity, newcapacity))
        else:  # free chunk is at the end
            self.chunks[-1].end = newcapacity

        self.capacity = newcapacity

    def free(self, offset, size):
        nch = Chunk(offset, offset + size)
        # insert sorted
        if offset > self.chunks[-1].start:  # new chuck at the end
            self.chunks.append(nch)
        else:  # new chuck needs to be inserted
            for ic, ch in enumerate(self.chunks):
                if offset <= ch.start:
                    self.chunks.insert(ic, nch)
                    break
        # merge chunks
        pch = self.chunks[0]
        newchunks = [pch]
        for ch in self.chunks[1:]:
            if pch.overlaps(ch):
                pch.merge(ch)
            else:
                newchunks.append(ch)
                pch = ch
        self.chunks = newchunks

    @abstractmethod
    def _new_buffer(self, capacity):
        "return newbuffer"

    @abstractmethod
    def update_from_native(self, offset, source, source_offset, nbytes):
        """Copy data from native buffer into self.buffer starting from offset"""

    @abstractmethod
    def copy_to_native(self, dest, dest_offset, source_offset, nbytes):
        """copy data from self.buffer into dest"""

    @abstractmethod
    def copy_native(self, offset, nbytes):
        """return native data with content at from offset and nbytes"""

    @abstractmethod
    def update_from_buffer(self, offset, source):
        """Copy data from python buffer such as bytearray, bytes, memoryview, numpy array.data"""

    @abstractmethod
    def to_nplike(self, offset, dtype, shape):
        """view in nplike"""

    @abstractmethod
    def update_from_nplike(self, offset, dest_dtype, value):
        """update data from nplike matching dest_dtype"""

    @abstractmethod
    def to_bytearray(self, offset, nbytes):
        """copy in byte array: used in update_from_xbuffer"""

    @abstractmethod
    def to_pointer_arg(self, offset, nbytes):
        """return data that can be used as argument in kernel"""

    def update_from_xbuffer(self, offset, source, source_offset, nbytes):
        """update from any xbuffer, don't pass through gpu if possible"""
        if source.context == self.context:
            self.update_from_native(
                offset, source.buffer, source_offset, nbytes
            )
        else:
            data = source.to_bytearray(source_offset, nbytes)
            self.update_from_buffer(offset, data)

    def get_free(self):
        return sum([ch.size for ch in self.chunks])

    def __repr__(self):
        name = self.__class__.__name__
        return f"<{name} {self.get_free()}/{self.capacity}>"


class Chunk:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    @property
    def size(self):
        return self.end - self.start

    #    def overlaps(self,other):
    #        return not ((other.end < self.start) or (other.start > self.end))

    def overlaps(self, other):
        return (other.end >= self.start) and (other.start <= self.end)

    def merge(self, other):
        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)
        return self

    def copy(self):
        return Chunk(self.start, self.end)

    def __repr__(self):
        return f"Chunk({self.start},{self.end})"


class View(NamedTuple):
    buffer: XBuffer
    offset: int
    size: int


available: List[Type[XContext]] = []


class Arg:
    def __init__(
        self, atype, pointer=False, name=None, const=False, factory=None
    ):
        self.atype = atype
        self.pointer = pointer
        self.name = name
        self.const = const
        self.factory = factory

    def get_c_type(self):
        ctype = self.atype._c_type
        if self.pointer:
            ctype += "*"
        return ctype


class Kernel:
    def __init__(self, args, c_name=None, ret=None, n_threads=1):
        self.c_name = c_name
        self.args = args
        self.ret = ret
        self.n_threads = n_threads

    def get_classes(self):
        classes = [
            a.atype for a in self.args if hasattr(a.atype, "_gen_c_api")
        ]
        if isinstance(self.ret, Arg) and hasattr(self.ret.atype, "_gen_c_api"):
            classes.append(self.ret.atype)
        return classes


class Method:
    def __init__(self, args, c_name, ret):
        self.args = args
        self.c_name = c_name
        self.ret = ret

    def mk_method(self):
        def a_method(instance, *args, **kwargs):
            context = instance._buffer.context
            kernel = context.kernels[self.kernel_name]
            kwargs[self.arg_name] = instance
            return kernel(*args, **kwargs)

        return a_method


def get_context_from_string(ctxstr):
    import xobjects as xo

    if ctxstr is None:
        return xo.ContexCPU()
    else:
        ll = ctxstr.split(":")
        if len(ll) <= 1:
            ctxtype = ll[0]
            option = []
        else:
            ctxtype, options = ctxstr.split(":")
            option = options.split(",")
    if ctxtype == "ContextCpu":
        if len(option) == 0:
            return xo.ContextCpu()
        else:
            return xo.ContextCpu(omp_num_threads=int(option[0]))
    elif ctxtype == "ContextCupy":
        if len(option) == 0:
            return xo.ContextCupy()
        else:
            return xo.ContextCupy(device=int(option[0]))
    elif ctxtype == "ContextPyopencl":
        if len(option) == 0:
            return xo.ContextPyopencl()
        else:
            return xo.ContextPyopencl(device=option[0])
    else:
        raise ValueError(f"Cannot create context from `{ctxstr}`")


def get_test_contexts():
    import os
    import xobjects as xo

    ctxstr = os.environ.get("XOBJECTS_TEST_CONTEXTS")
    if ctxstr is None:
        yield xo.ContextCpu()
        yield xo.ContextCpu(omp_num_threads=2)
        if xo.ContextCupy in xo.context.available:
            yield xo.ContextCupy()
        if xo.ContextPyopencl in xo.context.available:
            yield xo.ContextPyopencl()
    elif ctxstr == "all":
        yield xo.ContextCpu()
        yield xo.ContextCpu(omp_num_threads=2)
        if xo.ContextCupy in xo.context.available:
            yield xo.ContextCupy()
        if xo.ContextPyopencl in xo.context.available:
            for dd in xo.ContextPyopencl.get_devices():
                yield xo.ContextPyopencl(device=dd)

    else:
        for cc in ctxstr.split(";"):
            yield get_context_from_string(cc)


def get_user_context():
    """
    Get the context specfied by the enviroment variable XOBJECTS_USER_CONTEXT.
    If not present use ContextCpu().

    Examples:
       ContextPyopencl:0.0  -> ContextPyopencl(device="0.0")
       ContextPyopencl:1.0  -> ContextPyopencl(device="1.0")
       ContextPyopencl      -> ContextPyopencl()
       ContextCpu           -> ContextCpu()
       ContextCpu:2         -> ContextCpu(omp_num_threads=2)
       ContextCupy:0        -> ContextCupy(device=0)
    """
    import os

    ctxstr = os.environ.get("XOBJECTS_USER_CONTEXT")
    return get_context_from_string(ctxstr)



from setuptools import setup

setup(
    name="xobjects",
    version="0.1.7",
    description="In-memory serialization and code generator for CPU and GPU",
    author="Riccardo De Maria",
    author_email="riccardo.de.maria@cern.ch",
    url="https://github.com/rdemaria/cobjects",
    python_requires=">=3.7",
    setup_requires=[],
    install_requires=["numpy", "cffi"],
    packages=["xobjects"],
)



import json

import numpy as np
from .struct import Struct
from .typeutils import context_default


class _FieldOfDressed:
    def __init__(self, name, XoStruct):
        self.name = name
        self.isnplikearray = False

        fnames = [ff.name for ff in XoStruct._fields]
        if self.name in fnames:
            ftype = getattr(XoStruct, self.name).ftype
            if hasattr(ftype, "_itemtype"):  # is xo object
                if hasattr(ftype._itemtype, "_dtype"):  # valid nplike object
                    self.isnplikearray = True

    def __get__(self, container, ContainerType=None):
        if self.isnplikearray:
            if hasattr(container, "_lim_arrays_name"):
                lim = getattr(container, container._lim_arrays_name)
                return getattr(container._xobject, self.name).to_nplike()[:lim]
            else:
                return getattr(container._xobject, self.name).to_nplike()
        elif hasattr(container, "_dressed_" + self.name):
            return getattr(container, "_dressed_" + self.name)
        else:
            return getattr(container._xobject, self.name)

    def __set__(self, container, value):
        if self.isnplikearray:
            self.__get__(container=container)[:] = value
        elif hasattr(value, "_xobject"):  # value is a dressed xobject
            setattr(container, "_dressed_" + self.name, value)
            setattr(container._xobject, self.name, value._xobject)
            getattr(container, self.name)._xobject = getattr(
                container._xobject, self.name
            )
        else:
            self.content = None
            setattr(container._xobject, self.name, value)


def dress(XoStruct, rename={}):

    if hasattr(XoStruct, "_DressingClass"):
        raise ValueError("A Struct cannot be dressed multiple times")

    DressedXStruct = type(
        "Dressed" + XoStruct.__name__, (), {"XoStruct": XoStruct}
    )

    DressedXStruct._rename = rename

    for ff in ["_buffer", "_offset"]:
        setattr(DressedXStruct, ff, _FieldOfDressed(ff, XoStruct))

    pynames_list = []
    for ff in XoStruct._fields:
        fname = ff.name
        if fname in rename.keys():
            pyname = rename[fname]
        else:
            pyname = fname
        pynames_list.append(pyname)

        setattr(DressedXStruct, pyname, _FieldOfDressed(fname, XoStruct))

        DressedXStruct._fields = pynames_list

    XoStruct._DressingClass = DressedXStruct

    def _move_to(self, _context=None, _buffer=None, _offset=None):
        self._xobject = self._xobject.__class__(
            self._xobject, _context=_context, _buffer=_buffer, _offset=_offset
        )
        self._reinit_from_xobject(_xobject=self._xobject)

    def _reinit_from_xobject(self, _xobject):
        self._xobject = _xobject
        for ff in self.XoStruct._fields:
            if hasattr(ff.ftype, "_DressingClass"):
                vv = ff.ftype._DressingClass(
                    _xobject=getattr(_xobject, ff.name)
                )
                pyname = self._rename.get(ff.name, ff.name)
                setattr(self, pyname, vv)

    def xoinitialize(self, _xobject=None, **kwargs):

        if _xobject is not None:
            self._reinit_from_xobject(_xobject=_xobject)
        else:
            # Handle dressed inputs
            dressed_kwargs = {}
            for kk, vv in kwargs.items():
                if hasattr(vv, "_xobject"):  # vv is dressed
                    dressed_kwargs[kk] = vv
                    kwargs[kk] = vv._xobject

            self._xobject = self.XoStruct(**kwargs)

            # Handle dressed inputs
            for kk, vv in dressed_kwargs.items():
                if kk in self._rename.keys():
                    pyname = self._rename[kk]
                else:
                    pyname = kk
                setattr(self, pyname, vv)

    def myinit(self, _xobject=None, **kwargs):
        self.xoinitialize(_xobject=_xobject, **kwargs)

    def to_dict(self, copy_to_cpu=True):
        out = {"__class__": self.__class__.__name__}
        # Use a cpu copy by default:
        if copy_to_cpu:
            obj = self.copy(_context=context_default)
        else:
            obj = self

        for ff in obj._fields:
            vv = getattr(obj, ff)
            if hasattr(vv, "to_dict"):
                out[ff] = vv.to_dict()
            elif hasattr(vv, "_to_dict"):
                out[ff] = vv._to_dict()
            else:
                out[ff] = vv
        return out

    @classmethod
    def from_dict(cls, dct, _context=None, _buffer=None, _offset=None):
        return cls(**dct, _context=_context, _buffer=_buffer, _offset=_offset)

    def copy(self, _context=None, _buffer=None, _offset=None):
        if _context is None and _buffer is None:
            _context = self._xobject._buffer.context
        # This makes a copy of the xobject
        xobject = self.XoStruct(
            self._xobject, _context=_context, _buffer=_buffer, _offset=_offset
        )
        return self.__class__(_xobject=xobject)

    def compile_custom_kernels(self, only_if_needed=False):
        context = self._buffer.context

        if only_if_needed:
            all_found = True
            for kk in self.XoStruct.custom_kernels.keys():
                if kk not in context.kernels.keys():
                    all_found = False
                    break
            if all_found:
                return

        context.add_kernels(
            sources=self.XoStruct.extra_sources,
            kernels=self.XoStruct.custom_kernels,
            extra_classes=[self.XoStruct],
            save_source_as="temp.c",
        )

    DressedXStruct.xoinitialize = xoinitialize
    DressedXStruct.compile_custom_kernels = compile_custom_kernels
    DressedXStruct.to_dict = to_dict
    DressedXStruct.from_dict = from_dict
    DressedXStruct.copy = copy
    DressedXStruct.__init__ = myinit
    DressedXStruct._reinit_from_xobject = _reinit_from_xobject
    DressedXStruct._move_to = _move_to

    return DressedXStruct


class JEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif np.issubdtype(type(obj), np.integer):
            return int(obj)
        else:
            return json.JSONEncoder.default(self, obj)


class MetaDressedStruct(type):
    def __new__(cls, name, bases, data):
        XoStruct_name = name + "Data"
        if "_xofields" in data.keys():
            xofields = data["_xofields"]
        else:
            for bb in bases:
                if hasattr(bb, "_xofields"):
                    xofields = bb._xofields
                    break
        XoStruct = type(XoStruct_name, (Struct,), xofields)

        bases = (dress(XoStruct),) + bases
        new_class = type.__new__(cls, name, bases, data)

        XoStruct._DressingClass = new_class

        return new_class


class DressedStruct(metaclass=MetaDressedStruct):
    _xofields = {}



import logging

import numpy as np

from .typeutils import Info, dispatch_arg, get_a_buffer, default_conf
from .scalar import Int64
from .array import Array

log = logging.getLogger(__name__)

NULLVALUE = -(2 ** 63)
NULLTYPE = -1
NULLREF = np.array([NULLVALUE, NULLTYPE], dtype="int64")

# Ref is a like a scalar
class MetaRef(type):
    def __getitem__(cls, reftype):
        return cls(reftype)


class Ref(metaclass=MetaRef):
    def __init__(self, reftype):
        self._reftype = reftype
        self.__name__ = "Ref" + self._reftype.__name__
        self._c_type = self.__name__

        self._size = 8

    def _from_buffer(self, buffer, offset):
        refoffset = Int64._from_buffer(buffer, offset)
        if refoffset == NULLVALUE:
            return None
        else:
            refoffset += offset  # from relative to absolute offset
            return self._reftype._from_buffer(buffer, refoffset)

    def _to_buffer(self, buffer, offset, value, info=None):

        # Get/set content
        if value is None:
            refoffset = NULLVALUE  # NULL value
        elif (
            value.__class__.__name__ == self._reftype.__name__  # same type
            and value._buffer is buffer
        ):
            refoffset = value._offset - offset

        else:
            newobj = self._reftype(value, _buffer=buffer)
            refoffset = newobj._offset - offset
        Int64._to_buffer(buffer, offset, refoffset)

    def __call__(self, *args):
        if len(args) == 0:
            return None
        else:
            (value,) = args
            return self._reftype(value)

    def _inspect_args(self, arg):
        return Info(size=self._size)

    def __getitem__(self, shape):
        return Array.mk_arrayclass(self, shape)

    def _gen_data_paths(self, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [self])
        if hasattr(self._reftype, "_gen_data_paths"):
            paths.extend(self._reftype._gen_data_paths(base + [self]))
        return paths

    def _gen_c_decl(self, conf=default_conf):
        from . import capi

        return capi.gen_cdefs(self, [], conf)

    def _gen_c_api(self, conf=default_conf):
        from . import capi

        return capi.gen_code(self, [], conf)

    def __repr__(self):
        return f"<ref {self.__name__}>"

    def _get_inner_types(self):
        return [self._reftype]


# UnionRef is a proper class because
#  - name generation is particularly inconvenient
#  - can be useful to instantiate for debugging
class MetaUnionRef(type):
    _reftypes: list
    _methods: list

    def __new__(cls, name, bases, data):
        if "_c_type" not in data:
            data["_c_type"] = name
        if "_methods" not in data:
            data["_methods"] = []

        return type.__new__(cls, name, bases, data)

    def _is_member(cls, value):
        typ = value.__class__
        for tt in cls._reftypes:
            if tt.__name__ == typ.__name__:
                return True
        return False

    def _typeid_from_type(cls, typ):
        for ii, tt in enumerate(cls._reftypes):
            if tt.__name__ == typ.__name__:
                return ii
        # If no match found:
        raise TypeError(f"{typ} is not memberof {cls}!")

    def _typeid_from_name(cls, name):
        for ii, tt in enumerate(cls._reftypes):
            if tt.__name__ == name:
                return ii
        # If no match found:
        raise TypeError(f"{name} is not memberof {cls}")

    def _type_from_name(cls, name):
        for tt in cls._reftypes:
            if tt.__name__ == name:
                return tt
        # If no match found:
        raise TypeError(f"{name} is not memberof {cls}")

    def _type_from_typeid(cls, typeid):
        for ii, tt in enumerate(cls._reftypes):
            if ii == typeid:
                return tt
        # If no match found:
        raise TypeError(f"Invalid id: {typeid}!")

    def _from_buffer(cls, buffer, offset):
        refoffset, typeid = Int64._array_from_buffer(buffer, offset, 2)
        if refoffset == NULLVALUE:
            return None
        else:
            reftype = cls._type_from_typeid(typeid)
            return reftype._from_buffer(buffer, offset + refoffset)

    def _inspect_args(cls, *args):
        """
        A unionref can be initialized with an instance of the classes in reftypes or
        a tuple (typename, dictionary)

        Input:
        - None
        - XObject
        - typename, dict
        """
        log.debug(f"get info for {cls} from {args}")
        info = Info(size=cls._size)
        return info

    def _to_buffer(cls, buffer, offset, value, info=None):
        if isinstance(value, cls):  # binary copy
            buffer.update_from_xbuffer(
                offset, value._buffer, value._offset, value._size
            )
        else:
            if value is None:
                xobj = None
            elif isinstance(value, tuple):
                if len(value) == 0:
                    xobj = None
                    typeid = None
                elif len(value) == 1:  # must be XObject or None
                    xobj = value[0]
                    if xobj is not None:
                        typ = xobj.__class__
                        typeid = cls._typeid_from_type(typ)
                        if xobj._buffer != buffer:
                            xobj = typ(xobj, _buffer=buffer)
                elif len(value) == 2:  # must be (str,dict)
                    tname, data = value
                    typ = cls._type_from_name(tname)
                    typeid = cls._typeid_from_name(tname)
                    xobj = typ(data, _buffer=buffer)
            elif cls._is_member(value):
                xobj = value
                typ = xobj.__class__
                typeid = cls._typeid_from_type(typ)
                if xobj._buffer != buffer:
                    xobj = typ(xobj, _buffer=buffer)
            else:
                raise ValueError(f"{value} not handled")
            if xobj is None:
                Int64._array_to_buffer(buffer, offset, NULLREF)
            else:
                ref = np.array([xobj._offset - offset, typeid])
                Int64._array_to_buffer(buffer, offset, ref)

    def __getitem__(cls, shape):
        return Array.mk_arrayclass(cls, shape)

    def _pre_init(cls, *arg, **kwargs):
        return kwargs

    def __repr__(cls):
        return f"<unionref {cls.__name__}>"


class UnionRef(metaclass=MetaUnionRef):
    _size = 16
    _reftypes: list

    def __init__(
        self, *args, _context=None, _buffer=None, _offset=None, **kwargs
    ):
        cls = self.__class__

        args, _ = self._pre_init(*args, **kwargs)

        self._buffer, self._offset = get_a_buffer(
            cls._size, _context, _buffer, _offset
        )

        cls._to_buffer(self._buffer, self._offset, args)

        self._post_init()

    def get(self):
        reloffset, typeid = Int64._array_from_buffer(
            self._buffer, self._offset, 2
        )
        if reloffset == NULLVALUE:
            return None
        else:
            cls = self.__class__
            typ = cls._type_from_typeid(typeid)
            offset = self._offset + reloffset
            return typ._from_buffer(self._buffer, offset)

    @classmethod
    def _pre_init(cls, *args, **kwargs):
        return args, kwargs

    def _post_init(self):
        pass

    @classmethod
    def _gen_data_paths(cls, base=None):
        paths = []
        if base is None:
            base = []
        paths.append(base + [cls])
        return paths

    @classmethod
    def _gen_c_decl(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_cdefs(cls, paths, conf)

    @classmethod
    def _gen_c_api(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_code(cls, paths, conf)

    @classmethod
    def _gen_kernels(cls, conf=default_conf):
        from . import capi

        paths = cls._gen_data_paths()
        return capi.gen_kernels(cls, paths, conf)

    @classmethod
    def _get_inner_types(cls):
        return cls._reftypes


def is_ref(atype):
    return isinstance(atype, Ref)


def is_unionref(atype):
    return isinstance(atype, MetaUnionRef)



import os
import logging

import numpy as np

from .context import (
    XBuffer,
    XContext,
    ModuleNotAvailable,
    available,
    classes_from_kernels,
    sort_classes,
    sources_from_classes,
    _concatenate_sources,
)

from .specialize_source import specialize_source
from .linkedarray import BaseLinkedArray

log = logging.getLogger(__name__)

try:
    import cupy
    from cupyx.scipy import fftpack as cufftp

    _enabled = True
except ImportError:
    log.info("cupy is not installed, ContextCupy will not be available")
    cupy = ModuleNotAvailable(
        message=("cupy is not installed. " "ContextCupy is not available!")
    )
    cufftp = cupy
    _enabled = False

if _enabled:
    # order of base classes matters as it defines which __setitem__ is used
    class LinkedArrayCupy(BaseLinkedArray, cupy.ndarray):

        @classmethod
        def _build_view(cls, a):
            assert len(a.shape) == 1
            return cls(shape=a.shape, dtype=a.dtype, memptr=a.data,
                    strides=a.strides, order='C')


cudaheader = [
    """\
typedef signed long long int64_t;  //only_for_context cuda
typedef signed char      int8_t;   //only_for_context cuda
typedef unsigned int     uint32_t; //only_for_context cuda

"""
]


def nplike_to_cupy(arr):
    return cupy.array(arr)


class ContextCupy(XContext):

    """
    Creates a Cupy Context object, that allows performing the computations
    on nVidia GPUs.

    Args:
        default_block_size (int):  CUDA thread size that is used by default
            for kernel execution in case a block size is not specified
            directly in the kernel object. The default value is 256.
        device (int): Identifier of the device to be used by the context.
    Returns:
        ContextCupy: context object.

    """

    @property
    def nplike_array_type(self):
        return cupy.ndarray

    @property
    def linked_array_type(self):
        return LinkedArrayCupy

    def __init__(self, default_block_size=256, device=None):

        if device is not None:
            cupy.cuda.Device(device).use()

        super().__init__()

        self.default_block_size = default_block_size

    def _make_buffer(self, capacity):
        return BufferCupy(capacity=capacity, context=self)

    def add_kernels(
        self,
        sources=[],
        kernels=[],
        specialize=True,
        save_source_as=None,
        extra_cdef=None,
        extra_classes=[],
        extra_headers=[],
    ):

        """
        Adds user-defined kernels to to the context. The kernel source
        code is provided as a string and/or in source files and must contain
        the kernel names defined in the kernel descriptions.
        Args:
            sources (list): List of source codes that are concatenated before
                compilation. The list can contain strings (raw source code),
                File objects and Path objects.
            kernels (dict): Dictionary with the kernel descriptions
                in the form given by the following examples. The descriptions
                define the kernel names, the type and name of the arguments
                and identify one input argument that defines the number of
                threads to be launched (only on cuda/opencl).
            specialize_code (bool): If True, the code is specialized using
                annotations in the source code. Default is ``True``
            save_source_as (str): Filename for saving the specialized source
                code. Default is ```None```.
        Example:

        .. code-block:: python

            # A simple kernel
            src_code = '''
            /*gpukern*/
            void my_mul(const int n,
                /*gpuglmem*/ const double* x1,
                /*gpuglmem*/ const double* x2,
                /*gpuglmem*/       double* y) {
                int tid = 0 //vectorize_over tid
                y[tid] = x1[tid] * x2[tid];
                //end_vectorize
                }
            '''

            # Prepare description
            kernel_descriptions = {
                "my_mul": xo.Kernel(
                    args=[
                        xo.Arg(xo.Int32, name="n"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x1"),
                        xo.Arg(xo.Float64, pointer=True, const=True, name="x2"),
                        xo.Arg(xo.Float64, pointer=True, const=False, name="y"),
                    ],
                    n_threads="n",
                    ),
            }

            # Import kernel in context
            ctx.add_kernels(
                sources=[src_code],
                kernels=kernel_descriptions,
                save_source_as=None,
            )

            # With a1, a2, b being arrays on the context, the kernel
            # can be called as follows:
            ctx.kernels.my_mul(n=len(a1), x1=a1, x2=a2, y=b)
        """

        classes = classes_from_kernels(kernels)
        classes.update(extra_classes)
        classes = sort_classes(classes)
        cls_sources = sources_from_classes(classes)

        headers = cudaheader + extra_headers

        sources = headers + cls_sources + sources

        source, folders = _concatenate_sources(sources)
        source = "\n".join(['extern "C"{', source, "}"])

        if specialize:
            # included files are searched in the same folders od the src_filed
            specialized_source = specialize_source(
                source, specialize_for="cuda", search_in_folders=folders
            )
        else:
            specialized_source = source

        if save_source_as is not None:
            with open(save_source_as, "w") as fid:
                fid.write(specialized_source)

        module = cupy.RawModule(code=specialized_source)

        for pyname, kernel in kernels.items():
            if kernel.c_name is None:
                kernel.c_name = pyname

            self.kernels[pyname] = KernelCupy(
                function=module.get_function(kernel.c_name),
                description=kernel,
                block_size=self.default_block_size,
                context=self,
            )

            self.kernels[pyname].source = source
            self.kernels[pyname].specialized_source = specialized_source

    def nparray_to_context_array(self, arr):
        """
        Copies a numpy array to the device memory.

        Args:
            arr (numpy.ndarray): Array to be transferred

        Returns:
            cupy.ndarray:The same array copied to the device.

        """
        dev_arr = cupy.array(arr)
        return dev_arr

    def nparray_from_context_array(self, dev_arr):
        """
        Copies an array to the device to a numpy array.

        Args:
            dev_arr (cupy.ndarray): Array to be transferred.
        Returns:
            numpy.ndarray: The same data copied to a numpy array.

        """
        return dev_arr.get()

    @property
    def nplike_lib(self):
        """
        Module containing all the numpy features supported by cupy.
        """
        return cupy

    def synchronize(self):
        """
        Ensures that all computations submitted to the context are completed.
        Equivalent to ``cupy.cuda.stream.get_current_stream().synchronize()``
        """
        cupy.cuda.stream.get_current_stream().synchronize()

    def zeros(self, *args, **kwargs):
        """
        Allocates an array of zeros on the device. The function has the same
        interface of numpy.zeros"""
        return self.nplike_lib.zeros(*args, **kwargs)

    def plan_FFT(
        self,
        data,
        axes,
    ):
        """
        Generates an FFT plan object to be executed on the context.

        Args:
            data (cupy.ndarray): Array having type and shape for which the FFT
                needs to be planned.
            axes (sequence of ints): Axes along which the FFT needs to be
                performed.
        Returns:
            FFTCupy: FFT plan for the required array shape, type and axes.

        Example:

        .. code-block:: python

            plan = context.plan_FFT(data, axes=(0,1))

            data2 = 2*data

            # Forward tranform (in place)
            plan.transform(data2)

            # Inverse tranform (in place)
            plan.itransform(data2)
        """
        return FFTCupy(self, data, axes)

    @property
    def kernels(self):

        """
        Dictionary containing all the kernels that have been imported to the context.
        The syntax ``context.kernels.mykernel`` can also be used.
        """

        return self._kernels


class BufferCupy(XBuffer):
    def _make_context(self):
        return ContextCupy()

    def _new_buffer(self, capacity):
        return cupy.zeros(shape=(capacity,), dtype=cupy.uint8)

    def update_from_native(self, offset, source, source_offset, nbytes):
        """Copy data from native buffer into self.buffer starting from offset"""
        self.buffer[offset : offset + nbytes] = source[
            source_offset : source_offset + nbytes
        ]

    def copy_native(self, offset, nbytes):
        """Return a new cupy buffer with data from offset"""
        return self.buffer[offset : offset + nbytes].copy()

    def copy_to_native(self, dest, dest_offset, source_offset, nbytes):
        """copy data from self to source from offset and nbytes"""
        dest[dest_offset : dest_offset + nbytes] = self.buffer[
            source_offset : source_offset + nbytes
        ]

    def update_from_buffer(self, offset, source):
        """Copy data from python buffer such as bytearray, bytes, memoryview, numpy array.data"""
        nbytes = len(source)
        self.buffer[offset : offset + nbytes] = cupy.array(
            np.frombuffer(source, dtype=np.uint8)
        )

    def to_nplike(self, offset, dtype, shape):
        """view in nplike"""
        nbytes = np.prod(shape) * np.dtype(dtype).itemsize
        return (
            self.buffer[offset : offset + nbytes]
            .view(dtype=dtype)
            .reshape(*shape)
        )

    def update_from_nplike(self, offset, dest_dtype, value):
        if dest_dtype != value.dtype:
            value = value.astype(dtype=dest_dtype)  # make a copy
        src = value.view("int8")
        self.buffer[offset : offset + src.nbytes] = value.view("int8")

    def to_bytearray(self, offset, nbytes):
        """copy in byte array: used in update_from_xbuffer"""
        return self.buffer[offset : offset + nbytes].get().tobytes()

    def to_pointer_arg(self, offset, nbytes):
        """return data that can be used as argument in kernel"""
        return self.buffer[offset : offset + nbytes]


class KernelCupy(object):
    def __init__(
        self,
        function,
        description,
        block_size,
        context,
    ):

        self.function = function
        self.description = description
        self.block_size = block_size
        self.context = context

    def to_function_arg(self, arg, value):
        if arg.pointer:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                if hasattr(value, "dtype"):  # nparray
                    assert isinstance(value, cupy.ndarray)
                    return value.data
                elif hasattr(value, "_shape"):  # xobject array
                    raise NotImplementedError
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} "
                    f"of kernel {self.description.pyname}"
                )
        else:
            if hasattr(arg.atype, "_dtype"):  # it is numerical scalar
                return arg.atype(value)  # try to return a numpy scalar
            elif hasattr(arg.atype, "_size"):  # it is a compound xobject
                return value._buffer.buffer[value._offset :]
            else:
                raise ValueError(
                    f"Invalid value {value} for argument {arg.name} of kernel {self.description.pyname}"
                )

    @property
    def num_args(self):
        return len(self.description.args)

    def __call__(self, **kwargs):
        assert len(kwargs.keys()) == self.num_args
        arg_list = []
        for arg in self.description.args:
            vv = kwargs[arg.name]
            arg_list.append(self.to_function_arg(arg, vv))

        if isinstance(self.description.n_threads, str):
            n_threads = kwargs[self.description.n_threads]
        else:
            n_threads = self.description.n_threads

        grid_size = int(np.ceil(n_threads / self.block_size))
        self.function((grid_size,), (self.block_size,), arg_list)


class FFTCupy(object):
    def __init__(self, context, data, axes):

        self.context = context
        self.axes = axes

        assert len(data.shape) > max(axes)

        from cupyx.scipy import fftpack as cufftp

        if data.flags.f_contiguous:
            self._ax = [data.ndim - 1 - aa for aa in axes]
            _dat = data.T
            self.f_contiguous = True
        else:
            self._ax = axes
            _dat = data
            self.f_contiguous = False

        self._fftplan = cufftp.get_fft_plan(
            _dat, axes=self._ax, value_type="C2C"
        )

    def transform(self, data):
        if self.f_contiguous:
            _dat = data.T
        else:
            _dat = data
        _dat[:] = cufftp.fftn(_dat, axes=self._ax, plan=self._fftplan)[:]
        """The transform is done inplace"""

    def itransform(self, data):
        """The transform is done inplace"""
        if self.f_contiguous:
            _dat = data.T
        else:
            _dat = data
        _dat[:] = cufftp.ifftn(_dat, axes=self._ax, plan=self._fftplan)[:]


if _enabled:
    available.append(ContextCupy)
